# Copyright (c) NXAI GmbH and its affiliates 2024
# Korbininan Pöppel
import torch
from typing import Callable


def round_to_multiple(n, m=8):
    return ((n + m - 1) // m) * m


def conditional_decorator(condition, decorator):
    """A higher-order decorator that applies 'decorator' only if 'condition' is True."""

    def dummy_decorator(func):
        """A dummy decorator that does nothing."""
        return func

    if condition:
        # If condition is True, return the actual decorator
        return decorator
    else:
        # If condition is False, return the dummy decorator
        return dummy_decorator


class ParameterProxy:
    """
    This class helps keeping parameters in a specialized internal structure to be optimal for
    computation speed, while having a proxied version to be called externally that is backend-agnostic.
    It takes a module and a parameter name of a parameter in that module it represents.
    Via __setitem__ and __getitem__ the "external"
    """

    def __init__(
        self,
        module,
        parameter_name,
        internal_to_external: Callable[[torch.Tensor], torch.Tensor],
        external_to_internal: Callable[[torch.Tensor], torch.Tensor],
    ):
        self.module = module
        self.parameter_name = parameter_name
        self.internal_to_external = internal_to_external
        self.external_to_internal = external_to_internal

    def __getitem__(self, key):
        # Transform and then apply the slice to the external shape
        external_param = self.internal_to_external(getattr(self.module, self.parameter_name)).detach()
        return external_param[key]

    def __setitem__(self, key, value):
        # Apply the slice on the external shape, then transform back
        with torch.no_grad():
            external_param = self.internal_to_external(getattr(self.module, self.parameter_name))
            external_param[key] = value
            getattr(self.module, self.parameter_name).data = self.external_to_internal(external_param).contiguous()

    def clone(self):
        return self.internal_to_external(getattr(self.module, self.parameter_name)).clone()

    @property
    def shape(self):
        return self.internal_to_external(getattr(self.module, self.parameter_name)).shape

    @property
    def ndim(self):
        return self.internal_to_external(getattr(self.module, self.parameter_name)).ndim

    @property
    def grad(self):
        return self.internal_to_external(getattr(self.module, self.parameter_name).grad)

    def __getattr__(self, name: str):
        return getattr(getattr(self.module, self.parameter_name), name)



# Copyright (c) NXAI GmbH and its affiliates 2024
# Korbininan Pöppel
from dataclasses import dataclass
from typing import Optional
import torch
from ...components.ln import MultiHeadLayerNorm
from ...components.linear_headwise import (
    LinearHeadwiseExpand,
    LinearHeadwiseExpandConfig,
)
from ...components.conv import CausalConv1d, CausalConv1dConfig
from ...components.init import small_init_init_

from torch import nn
from .cell import sLSTMCell, sLSTMCellConfig


@dataclass
class sLSTMLayerConfig(sLSTMCellConfig):
    embedding_dim: int = -1
    num_heads: int = (
        4  # this must divide the hidden size, is not yet supported by all versions in this directory
    )
    conv1d_kernel_size: int = 4  # 0 means no convolution included
    group_norm_weight: bool = True
    dropout: float = 0.0

    def __post_init__(self):
        self.hidden_size = self.embedding_dim
        sLSTMCellConfig.__post_init__(self)


class sLSTMLayer(nn.Module):
    config_class = sLSTMLayerConfig

    def __init__(self, config: sLSTMLayerConfig):
        super().__init__()
        self.config = config

        if self.config.conv1d_kernel_size > 0:
            self.conv1d = CausalConv1d(
                config=CausalConv1dConfig(
                    feature_dim=self.config.embedding_dim,
                    kernel_size=self.config.conv1d_kernel_size,
                )
            )
            self.conv_act_fn = nn.SiLU()

        self.fgate = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config.embedding_dim,
                num_heads=self.config.num_heads,
                bias=False,
            )
        )
        self.igate = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config.embedding_dim,
                num_heads=self.config.num_heads,
                bias=False,
            )
        )
        self.zgate = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config.embedding_dim,
                num_heads=self.config.num_heads,
                bias=False,
            )
        )
        self.ogate = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config.embedding_dim,
                num_heads=self.config.num_heads,
                bias=False,
            )
        )

        self.slstm_cell = sLSTMCell(self.config)
        self.group_norm = MultiHeadLayerNorm(
            ndim=self.config.embedding_dim, weight=self.config.group_norm_weight
        )
        self.dropout = nn.Dropout(self.config.dropout)

    def reset_parameters(self):
        self.slstm_cell.reset_parameters()
        self.group_norm.reset_parameters()
        small_init_init_(self.igate.weight, dim=self.config.embedding_dim)
        small_init_init_(self.fgate.weight, dim=self.config.embedding_dim)
        small_init_init_(self.zgate.weight, dim=self.config.embedding_dim)
        small_init_init_(self.ogate.weight, dim=self.config.embedding_dim)

    def step(
        self,
        x: torch.Tensor,
        conv_state: Optional[torch.Tensor] = None,
        slstm_state: Optional[torch.Tensor] = None,
    ):
        B, S, _ = x.shape

        if self.config.conv1d_kernel_size > 0:
            x_conv, conv_state = self.conv1d.step(x, conv_state=conv_state)
            x_conv = self.conv_act_fn(x_conv)
        else:
            x_conv = x

        i, f, z, o = (
            self.fgate(x_conv),
            self.igate(x_conv),
            self.zgate(x),
            self.ogate(x),
        )

        y, slstm_state = self.slstm_cell(
            torch.cat([i, f, z, o], dim=-1), state=slstm_state
        )

        y = self.dropout(y)

        out = self.group_norm(y).transpose(1, 2).view(B, S, -1)

        return out, {"conv_state": conv_state, "slstm_state": slstm_state}

    def forward(
        self,
        x: torch.Tensor,
        conv_state: Optional[torch.Tensor] = None,
        slstm_state: Optional[torch.Tensor] = None,
        return_last_state=False,
        **kwargs,
    ) -> torch.Tensor:
        B, S, _ = x.shape

        if self.config.conv1d_kernel_size > 0:
            if return_last_state:
                x_conv, conv_state = self.conv1d(
                    x, conv_state, return_last_state=return_last_state
                )
            else:
                x_conv = self.conv1d(x, conv_state, return_last_state=return_last_state)
            x_conv = self.conv_act_fn(x_conv)
        else:
            x_conv = x

        i, f, z, o = (
            self.fgate(x_conv),
            self.igate(x_conv),
            self.zgate(x),
            self.ogate(x),
        )

        y, slstm_state = self.slstm_cell(
            torch.cat([i, f, z, o], dim=-1), state=slstm_state
        )

        y = self.dropout(y)

        out = self.group_norm(y).transpose(1, 2).view(B, S, -1)

        if return_last_state:
            return out, {"conv_state": conv_state, "slstm_state": slstm_state}
        else:
            return out



# Copyright (c) NXAI GmbH and its affiliates 2023
# Korbinian Poeppel

import os
from typing import Sequence, Union
import logging

import time
import random

import torch
from torch.utils.cpp_extension import load as _load

LOGGER = logging.getLogger(__name__)


def defines_to_cflags(defines=Union[dict[str, Union[int, str]], Sequence[tuple[str, Union[str, int]]]]):
    cflags = []
    print(defines)
    if isinstance(defines, dict):
        defines = defines.items()
    for key, val in defines:
        cflags.append(f"-D{key}={str(val)}")
    return cflags


curdir = os.path.dirname(__file__)

if torch.cuda.is_available():
    os.environ["CUDA_LIB"] = os.path.join(os.path.split(torch.utils.cpp_extension.include_paths(cuda=True)[-1])[0], "lib")


def load(*, name, sources, extra_cflags=(), extra_cuda_cflags=(), **kwargs):
    suffix = ""
    for flag in extra_cflags:
        pref = [st[0] for st in flag[2:].split("=")[0].split("_")]
        if len(pref) > 1:
            pref = pref[1:]
        suffix += "".join(pref)
        value = flag[2:].split("=")[1].replace("-", "m").replace(".", "d")
        value_map = {"float": "f", "__half": "h", "__nv_bfloat16": "b", "true": "1", "false": "0"}
        if value in value_map:
            value = value_map[value]
        suffix += value
    if suffix:
        suffix = "_" + suffix
    suffix = suffix[:64]

    extra_cflags = list(extra_cflags) + [
        "-U__CUDA_NO_HALF_OPERATORS__",
        "-U__CUDA_NO_HALF_CONVERSIONS__",
        "-U__CUDA_NO_BFLOAT16_OPERATORS__",
        "-U__CUDA_NO_BFLOAT16_CONVERSIONS__",
        "-U__CUDA_NO_BFLOAT162_OPERATORS__",
        "-U__CUDA_NO_BFLOAT162_CONVERSIONS__",
    ]

    myargs = {
        "verbose": True,
        "with_cuda": True,
        "extra_ldflags": [f"-L{os.environ['CUDA_LIB']}", "-lcublas"],
        "extra_cflags": [*extra_cflags],
        "extra_cuda_cflags": [
            # "-gencode",
            # "arch=compute_70,code=compute_70",
            # "-dbg=1",
            '-Xptxas="-v"',
            "-gencode",
            "arch=compute_80,code=compute_80",
            "-res-usage",
            "--use_fast_math",
            "-O3",
            "-Xptxas -O3",
            "--extra-device-vectorization",
            *extra_cflags,
            *extra_cuda_cflags,
        ],
    }
    print(myargs)
    myargs.update(**kwargs)
    # add random waiting time to minimize deadlocks because of badly managed multicompile of pytorch ext
    time.sleep(random.random() * 10)
    LOGGER.info(f"Before compilation and loading of {name}.")
    mod = _load(name + suffix, sources, **myargs)
    LOGGER.info(f"After compilation and loading of {name}.")
    return mod



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from dataclasses import dataclass
from typing import Sequence

import torch
from torch import nn

from .components.init import small_init_init_
from .utils import WeightDecayOptimGroupMixin
from .xlstm_block_stack import xLSTMBlockStack, xLSTMBlockStackConfig


@dataclass
class xLSTMLMModelConfig(xLSTMBlockStackConfig):
    vocab_size: int = -1
    tie_weights: bool = False
    weight_decay_on_embedding: bool = False
    add_embedding_dropout: bool = False


class xLSTMLMModel(WeightDecayOptimGroupMixin, nn.Module):
    config_class = xLSTMLMModelConfig

    def __init__(self, config: xLSTMLMModelConfig, **kwargs):
        super().__init__()
        self.config = config

        self.xlstm_block_stack = xLSTMBlockStack(config=config)
        self.token_embedding = nn.Embedding(num_embeddings=config.vocab_size, embedding_dim=config.embedding_dim)
        self.emb_dropout = nn.Dropout(config.dropout) if config.add_embedding_dropout else nn.Identity()

        self.lm_head = nn.Linear(
            in_features=config.embedding_dim,
            out_features=config.vocab_size,
            bias=False,
        )
        if config.tie_weights:
            self.lm_head.weight = self.token_embedding.weight

    def reset_parameters(self):
        self.xlstm_block_stack.reset_parameters()

        small_init_init_(self.token_embedding.weight, dim=self.config.embedding_dim)

        if not self.config.tie_weights:
            small_init_init_(self.lm_head.weight, dim=self.config.embedding_dim)

    def forward(self, idx: torch.Tensor) -> torch.Tensor:
        x = self.token_embedding(idx)
        x = self.emb_dropout(x)
        x = self.xlstm_block_stack(x)
        logits = self.lm_head(x)
        return logits

    def step(
        self, idx: torch.Tensor, state: dict[str, dict[str, tuple[torch.Tensor, ...]]] = None, **kwargs
    ) -> tuple[torch.Tensor, dict[str, dict[str, tuple[torch.Tensor, ...]]]]:
        x = self.token_embedding(idx)
        x = self.emb_dropout(x)
        x, state = self.xlstm_block_stack.step(x, state=state, **kwargs)
        logits = self.lm_head(x)
        return logits, state

    def _create_weight_decay_optim_groups(self, **kwargs) -> tuple[Sequence[nn.Parameter], Sequence[nn.Parameter]]:
        weight_decay, no_weight_decay = super()._create_weight_decay_optim_groups(**kwargs)
        # remove token embedding and add it to the correct group, accrording to the config
        weight_decay = list(weight_decay)
        removed = 0
        for idx in range(len(weight_decay)):
            if weight_decay[idx - removed] is self.token_embedding.weight:
                weight_decay.pop(idx - removed)
                removed += 1
        weight_decay = tuple(weight_decay)
        if self.config.weight_decay_on_embedding:
            weight_decay += (self.token_embedding.weight,)
        else:
            no_weight_decay += (self.token_embedding.weight,)

        return weight_decay, no_weight_decay



import torch
import pytest
from .template_chunkwise_vs_recurrent import template_test_recurrent_vs_chunkwise


def test_recurrent_vs_chunkwise_native(
    chunkwise_kernel_name: str = "chunkwise--native_autograd",
    sequence_kernel_name: str = "native_sequence__native",
    step_kernel_name: str = "native",
):
    template_test_recurrent_vs_chunkwise(
        chunkwise_kernel_name=chunkwise_kernel_name,
        sequence_kernel_name=sequence_kernel_name,
        step_kernel_name=step_kernel_name,
    )


@pytest.mark.parametrize(
    "chunkwise_kernel_name",
    [
        "chunkwise--native_autograd",
        "chunkwise--triton_xl_chunk",
        "chunkwise--triton_limit_chunk",
    ],
)
@pytest.mark.parametrize(
    "sequence_kernel_name, step_kernel_name",
    [
        ["native_sequence__native", "native"],
        ["native_sequence__triton_step_fused", "triton_fused"],
    ],
)
def test_recurrent_vs_chunkwise_triton(
    chunkwise_kernel_name: str,
    sequence_kernel_name: str,
    step_kernel_name: str,
):
    template_test_recurrent_vs_chunkwise(
        chunkwise_kernel_name=chunkwise_kernel_name,
        sequence_kernel_name=sequence_kernel_name,
        step_kernel_name=step_kernel_name,
    )



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from dataclasses import dataclass

import torch
from torch import nn

from ...components.conv import CausalConv1d, CausalConv1dConfig
from ...components.init import small_init_init_, wang_init_
from ...components.linear_headwise import (
    LinearHeadwiseExpand,
    LinearHeadwiseExpandConfig,
)
from ...utils import UpProjConfigMixin
from .cell import mLSTMCell, mLSTMCellConfig


@dataclass
class mLSTMLayerConfig(UpProjConfigMixin):
    conv1d_kernel_size: int = 4
    qkv_proj_blocksize: int = 4
    num_heads: int = 4
    proj_factor: float = 2.0

    # will be set toplevel config
    embedding_dim: int = -1
    bias: bool = False
    dropout: float = 0.0
    context_length: int = -1

    _num_blocks: int = 1
    _inner_embedding_dim: int = None

    def __post_init__(self):
        self._set_proj_up_dim(embedding_dim=self.embedding_dim)
        self._inner_embedding_dim = self._proj_up_dim


class mLSTMLayer(nn.Module):
    config_class = mLSTMLayerConfig

    def __init__(self, config: mLSTMLayerConfig):
        super().__init__()
        self.config = config

        self.proj_up = nn.Linear(
            in_features=self.config.embedding_dim,
            out_features=2 * self.config._inner_embedding_dim,
            bias=self.config.bias,
        )

        num_proj_heads = round(self.config._inner_embedding_dim // self.config.qkv_proj_blocksize)
        self.q_proj = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config._inner_embedding_dim,
                num_heads=num_proj_heads,
                bias=self.config.bias,
            )
        )
        self.k_proj = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config._inner_embedding_dim,
                num_heads=num_proj_heads,
                bias=self.config.bias,
            )
        )
        self.v_proj = LinearHeadwiseExpand(
            config=LinearHeadwiseExpandConfig(
                in_features=self.config._inner_embedding_dim,
                num_heads=num_proj_heads,
                bias=self.config.bias,
            )
        )

        self.conv1d = CausalConv1d(
            config=CausalConv1dConfig(
                feature_dim=self.config._inner_embedding_dim,
                kernel_size=self.config.conv1d_kernel_size,
            )
        )
        self.conv_act_fn = nn.SiLU()
        self.mlstm_cell = mLSTMCell(
            config=mLSTMCellConfig(
                context_length=self.config.context_length,
                embedding_dim=self.config._inner_embedding_dim,
                num_heads=self.config.num_heads,
            )
        )
        self.ogate_act_fn = nn.SiLU()

        self.learnable_skip = nn.Parameter(torch.ones(self.config._inner_embedding_dim, requires_grad=True))

        self.proj_down = nn.Linear(
            in_features=self.config._inner_embedding_dim,
            out_features=self.config.embedding_dim,
            bias=self.config.bias,
        )
        self.dropout = nn.Dropout(self.config.dropout)
        self.reset_parameters()

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        B, S, _ = x.shape

        # up-projection
        x_inner = self.proj_up(x)
        x_mlstm, z = torch.split(x_inner, split_size_or_sections=self.config._inner_embedding_dim, dim=-1)

        # mlstm branch
        x_mlstm_conv = self.conv1d(x_mlstm)
        x_mlstm_conv_act = self.conv_act_fn(x_mlstm_conv)

        q = self.q_proj(x_mlstm_conv_act)
        k = self.k_proj(x_mlstm_conv_act)
        v = self.v_proj(x_mlstm)

        h_tilde_state = self.mlstm_cell(q=q, k=k, v=v)

        h_tilde_state_skip = h_tilde_state + (self.learnable_skip * x_mlstm_conv_act)

        # output / z branch
        h_state = h_tilde_state_skip * self.ogate_act_fn(z)

        # down-projection
        y = self.dropout(self.proj_down(h_state))
        return y

    def step(
        self,
        x: torch.Tensor,
        mlstm_state: tuple[torch.Tensor, torch.Tensor, torch.Tensor] = None,
        conv_state: tuple[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, dict[str, tuple[torch.Tensor, ...]]]:
        B, S, _ = x.shape

        # up-projection
        x_inner = self.proj_up(x)
        x_mlstm, z = torch.split(x_inner, split_size_or_sections=self.config._inner_embedding_dim, dim=-1)

        # mlstm branch
        x_mlstm_conv, conv_state = self.conv1d.step(x_mlstm, conv_state=conv_state)
        x_mlstm_conv_act = self.conv_act_fn(x_mlstm_conv)

        q = self.q_proj(x_mlstm_conv_act)
        k = self.k_proj(x_mlstm_conv_act)
        v = self.v_proj(x_mlstm)

        h_tilde_state, mlstm_state = self.mlstm_cell.step(q=q, k=k, v=v, mlstm_state=mlstm_state)

        h_tilde_state_skip = h_tilde_state + (self.learnable_skip * x_mlstm_conv_act)

        # output / z branch
        h_state = h_tilde_state_skip * self.ogate_act_fn(z)

        # down-projection
        y = self.dropout(self.proj_down(h_state))
        return y, {"mlstm_state": mlstm_state, "conv_state": conv_state}

    def reset_parameters(self):
        # init inproj
        small_init_init_(self.proj_up.weight, dim=self.config.embedding_dim)
        if self.proj_up.bias is not None:
            nn.init.zeros_(self.proj_up.bias)
        # init outproj
        wang_init_(self.proj_down.weight, dim=self.config.embedding_dim, num_blocks=self.config._num_blocks)
        if self.proj_down.bias is not None:
            nn.init.zeros_(self.proj_down.bias)

        nn.init.ones_(self.learnable_skip)

        def _init_qkv_proj(qkv_proj: LinearHeadwiseExpand):
            # use the embedding dim instead of the inner embedding dim
            small_init_init_(qkv_proj.weight, dim=self.config.embedding_dim)
            if qkv_proj.bias is not None:
                nn.init.zeros_(qkv_proj.bias)

        _init_qkv_proj(self.q_proj)
        _init_qkv_proj(self.k_proj)
        _init_qkv_proj(self.v_proj)

        self.mlstm_cell.reset_parameters()



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from dataclasses import dataclass

import torch
from torch import nn

from ...components.init import bias_linspace_init_
from ...components.ln import MultiHeadLayerNorm
from .backends import parallel_stabilized_simple, recurrent_step_stabilized_simple


@dataclass
class mLSTMCellConfig:
    context_length: int = -1
    embedding_dim: int = -1
    num_heads: int = -1


class mLSTMCell(nn.Module):
    config_class = mLSTMCellConfig

    def __init__(self, config: mLSTMCellConfig):
        super().__init__()
        self.config = config

        self.backend_fn = parallel_stabilized_simple
        self.backend_fn_step = recurrent_step_stabilized_simple

        self.igate = nn.Linear(3 * config.embedding_dim, config.num_heads)
        self.fgate = nn.Linear(3 * config.embedding_dim, config.num_heads)

        self.outnorm = MultiHeadLayerNorm(ndim=config.embedding_dim, weight=True, bias=False)

        self.register_buffer(
            "causal_mask",
            torch.tril(torch.ones(config.context_length, config.context_length, dtype=torch.bool)),
            persistent=False,
        )

        self.reset_parameters()

    def forward(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, **kwargs) -> torch.Tensor:
        B, S, _ = q.shape  # (B, S, H)

        if_gate_input = torch.cat([q, k, v], dim=-1)
        q = q.view(B, S, self.config.num_heads, -1)  # (B, S, NH, DH)
        k = k.view(B, S, self.config.num_heads, -1)  # (B, S, NH, DH)
        v = v.view(B, S, self.config.num_heads, -1)  # (B, S, NH, DH)

        q = q.transpose(1, 2)  # (B, NH, S, DH)
        k = k.transpose(1, 2)  # (B, NH, S, DH)
        v = v.transpose(1, 2)  # (B, NH, S, DH)

        # compute input and forget gate pre-activations
        igate_preact = self.igate(if_gate_input)  # (B, S, NH)
        igate_preact = igate_preact.transpose(-1, -2).unsqueeze(-1)  # (B, NH, S, 1)
        fgate_preact = self.fgate(if_gate_input)  # (B, S, NH)
        fgate_preact = fgate_preact.transpose(-1, -2).unsqueeze(-1)  # (B, NH, S, 1)#

        h_state = self.backend_fn(
            queries=q,
            keys=k,
            values=v,
            igate_preact=igate_preact,
            fgate_preact=fgate_preact,
            lower_triangular_matrix=self.causal_mask,
        )  # (B, NH, S, DH)

        h_state_norm = self.outnorm(h_state)  # (B, NH, S, DH)
        h_state_norm = h_state_norm.transpose(1, 2).reshape(B, S, -1)  # (B, NH, S, DH) -> (B, S, NH, DH) -> (B, S, H)

        return h_state_norm

    def step(
        self,
        q: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        mlstm_state: tuple[torch.Tensor, torch.Tensor, torch.Tensor] = None,
        **kwargs,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        B, S, _ = q.shape  # (B, S, H)
        assert S == 1, f"mLSTMCell.step only supports sequence length S=1, but got S={S}."

        if_gate_input = torch.cat([q, k, v], dim=-1)
        q = q.view(B, S, self.config.num_heads, -1)  # (B, S, NH, DH)
        k = k.view(B, S, self.config.num_heads, -1)  # (B, S, NH, DH)
        v = v.view(B, S, self.config.num_heads, -1)  # (B, S, NH, DH)

        _, _, NH, DH = q.shape

        q = q.transpose(1, 2)  # (B, NH, S, DH)
        k = k.transpose(1, 2)  # (B, NH, S, DH)
        v = v.transpose(1, 2)  # (B, NH, S, DH)

        # compute input and forget gate pre-activations
        igate_preact = self.igate(if_gate_input)  # (B, S, NH)
        igate_preact = igate_preact.transpose(-1, -2).unsqueeze(-1)  # (B, NH, S, 1)
        fgate_preact = self.fgate(if_gate_input)  # (B, S, NH)
        fgate_preact = fgate_preact.transpose(-1, -2).unsqueeze(-1)  # (B, NH, S, 1)

        if mlstm_state is None:
            c_state = torch.zeros(size=(B, NH, DH, DH), device=q.device, dtype=q.dtype)
            n_state = torch.zeros(size=(B, NH, DH, 1), device=q.device, dtype=q.dtype)
            m_state = torch.zeros(size=(B, NH, 1, 1), device=q.device, dtype=q.dtype)
        else:
            c_state, n_state, m_state = mlstm_state
            c_state = c_state.to(device=q.device, dtype=q.dtype)
            n_state = n_state.to(device=q.device, dtype=q.dtype)
            m_state = m_state.to(device=q.device, dtype=q.dtype)

        assert c_state.shape == (B, NH, DH, DH), f"Expected c_state shape {(B, NH, DH, DH)}, but got {c_state.shape}."
        assert n_state.shape == (B, NH, DH, 1), f"Expected n_state shape {(B, NH, DH, 1)}, but got {n_state.shape}."
        assert m_state.shape == (B, NH, 1, 1), f"Expected m_state shape {(B, NH, 1, 1)}, but got {m_state.shape}."

        h_state, mlstm_state = self.backend_fn_step(
            c_state=c_state,
            n_state=n_state,
            m_state=m_state,
            q=q,
            k=k,
            v=v,
            igate_preact=igate_preact,
            fgate_preact=fgate_preact,
        )  # (B, NH, 1 DH), ((B, NH, DH, DH), (B, NH, DH, 1), (B, NH, 1, 1))

        h_state_norm = self.outnorm(h_state)  # (B, NH, S, DH)
        h_state_norm = h_state_norm.transpose(1, 2).reshape(B, S, -1)  # (B, NH, S, DH) -> (B, S, NH, DH) -> (B, S, H)

        return h_state_norm, mlstm_state

    def reset_parameters(self):
        self.outnorm.reset_parameters()
        # forget gate initialization
        torch.nn.init.zeros_(self.fgate.weight)
        bias_linspace_init_(self.fgate.bias, start=3.0, end=6.0)
        # input gate initialization
        torch.nn.init.zeros_(self.igate.weight)
        torch.nn.init.normal_(self.igate.bias, mean=0.0, std=0.1)



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck, Korbininan Pöppel
from dataclasses import dataclass

from math import sqrt
import torch

# from einops import einsum, rearrange
from torch import nn


@dataclass
class LinearHeadwiseExpandConfig:
    in_features: int = 0
    # this is the number of heads that the in_features are split into
    # if num_heads=1, this is a normal linear layer
    # if num_heads>1, the in_features are split into num_heads and each head is projected separately
    # if num_heads=in_features, each feature is projected separately
    num_heads: int = -1
    expand_factor_up: float = 1

    # this is internally computed
    # but can be overwritten if you want to use a different output dimension
    # if > 0 the expand factor is ignored
    _out_features: int = -1

    bias: bool = True
    trainable_weight: bool = True
    trainable_bias: bool = True

    def __post_init__(self):
        assert self.num_heads > 0, "num_heads must be set"
        assert self.num_heads <= self.in_features, "num_heads must be <= in_features"
        assert (
            self.in_features % self.num_heads == 0
        ), "in_features must be a multiple of num_heads"

        if self._out_features < 0:
            self._out_features = round(self.expand_factor_up * self.in_features)


class LinearHeadwiseExpand(nn.Module):
    """This is a structured projection layer that projects the input to a higher dimension.
    It only allows integer up-projection factors, i.e. the output dimension is a multiple of the input dimension.
    """

    config_class = LinearHeadwiseExpandConfig

    def __init__(self, config: LinearHeadwiseExpandConfig):
        super().__init__()
        self.config = config
        in_features = self.config.in_features
        num_heads = self.config.num_heads
        out_features_per_head = config._out_features // num_heads
        self.weight = nn.Parameter(
            torch.empty(num_heads, out_features_per_head, in_features // num_heads),
            requires_grad=config.trainable_weight,
        )
        if config.bias:
            self.bias = nn.Parameter(
                torch.empty(config._out_features), requires_grad=config.trainable_bias
            )
        else:
            self.bias = None
        self.reset_parameters()

    def reset_parameters(self, **kwargs):
        # small init
        nn.init.normal_(
            self.weight.data, mean=0.0, std=sqrt(2 / 5 / self.weight.shape[-1])
        )
        if self.bias is not None:
            nn.init.zeros_(self.bias.data)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shape = x.shape
        x = x.view(*shape[:-1], self.config.num_heads, -1)
        x = torch.einsum("...hd,hod->...ho", x, self.weight)
        x = x.reshape(*shape[:-1], -1)
        if self.bias is not None:
            x = x + self.bias
        return x

    def extra_repr(self):
        return (
            f"in_features={self.config.in_features}, "
            f"num_heads={self.config.num_heads}, "
            f"expand_factor_up={self.config.expand_factor_up}, "
            f"bias={self.config.bias}, "
            f"trainable_weight={self.config.trainable_weight}, "
            f"trainable_bias={self.config.trainable_bias}, "
        )



# Copyright (c) NXAI GmbH and its affiliates 2023
# Korbininan Pöppel

from typing import Callable
import torch

from .slstm import slstm_forward_pointwise as slstm_forward_pointwise_slstm
from .lstm import slstm_forward_pointwise as slstm_forward_pointwise_lstm


slstm_pointwise_function_registry: dict[str, Callable] = {
    "slstm": slstm_forward_pointwise_slstm,
    "lstm": slstm_forward_pointwise_lstm,
}


def slstm_forward(
    x: torch.Tensor,  # [S, B, G*I]
    states: torch.Tensor,  # [4, B, H] only the first is used for recurrence!
    R: torch.Tensor,  # [K, R*H, H] - K num_heads
    b: torch.Tensor,  # [T*H]
    pointwise_forward: Callable[
        tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, dict[str, float]],
        tuple[torch.Tensor, torch.Tensor],
    ],
    constants: dict[str, float] = {},
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    num_states = states.shape[0]
    sequence_dim = x.shape[0]
    num_gates_r = (
        R.shape[1] // R.shape[2]
    )  # this only works for a fully-connected RNN, for a hin change this
    hidden_dim = R.shape[2] * R.shape[0]
    num_gates_t = b.shape[0] // hidden_dim
    batch_dim = x.shape[1]
    num_heads = R.shape[0]
    head_dim = R.shape[2]

    assert batch_dim == states.shape[1]
    assert hidden_dim == states.shape[2]

    g = torch.zeros(
        [sequence_dim + 1, num_gates_t, batch_dim, hidden_dim],
        device=x.device,
        dtype=x.dtype,
    )

    states_all = torch.zeros(
        [num_states, sequence_dim + 1, batch_dim, hidden_dim],
        device=x.device,
        dtype=x.dtype,
    )
    states_all[:, 0] = states
    for i, Wx_t in enumerate(x.unbind(dim=0)):
        Ry = (
            states[0]
            .reshape(batch_dim, num_heads, 1, -1)
            .matmul(
                R.transpose(1, 2).reshape(
                    1, num_heads, head_dim, num_gates_r * head_dim
                )
            )
            .reshape(batch_dim, num_heads, num_gates_r, -1)
            .transpose(1, 2)
            .reshape(batch_dim, -1)
        )
        sdtype = states.dtype
        states, gates = pointwise_forward(Wx_t, Ry, b, states, constants=constants)
        states = states.to(dtype=sdtype)
        g[i] = gates
        states_all[:, i + 1] = states

    # shapes ([S, B, H], ([B,H], [B,H], [B,H]), [S, B, 4*H])
    return states_all, states, g


def slstm_forward_step(
    x: torch.Tensor,  # [B, G*I]
    states: torch.Tensor,  # [4, B, H] only the first is used for recurrence!
    R: torch.Tensor,  # [K, R*H, H] - K num_heads
    b: torch.Tensor,  # [T*H]
    pointwise_forward: Callable[
        tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, dict[str, float]],
        tuple[torch.Tensor, torch.Tensor],
    ],
    constants: dict[str, float] = {},
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    num_states = states.shape[0]
    sequence_dim = x.shape[0]
    num_gates_r = (
        R.shape[1] // R.shape[2]
    )  # this only works for a fully-connected RNN, for a hin change this
    hidden_dim = R.shape[2] * R.shape[0]
    num_gates_t = b.shape[0] // hidden_dim
    batch_dim = x.shape[1]
    num_heads = R.shape[0]
    head_dim = R.shape[2]

    assert batch_dim == states.shape[1]
    assert hidden_dim == states.shape[2]

    g = torch.zeros(
        [sequence_dim + 1, num_gates_t, batch_dim, hidden_dim],
        device=x.device,
        dtype=x.dtype,
    )

    states_all = torch.zeros(
        [num_states, sequence_dim + 1, batch_dim, hidden_dim],
        device=x.device,
        dtype=x.dtype,
    )
    states_all[:, 0] = states
    Ry = (
        states[0]
        .reshape(batch_dim, num_heads, 1, -1)
        .matmul(
            R.transpose(1, 2).reshape(1, num_heads, head_dim, num_gates_r * head_dim)
        )
        .reshape(batch_dim, num_heads, num_gates_r, -1)
        .transpose(1, 2)
        .reshape(batch_dim, -1)
    )
    sdtype = states.dtype
    states, gates = pointwise_forward(x[0], Ry, b, states, constants=constants)
    states = states.to(dtype=sdtype)

    # shapes ([S, B, H], ([B,H], [B,H], [B,H]), [S, B, 4*H])
    return states[:, None, ...], g[:, None, ...]



from dataclasses import dataclass, field

try:
    from mlstm_kernels.torch.backend_module import (
        mLSTMBackendConfig,
        mLSTMBackend,
        ChunkwiseKernelType,
        SequenceKernelType,
        StepKernelType,
        DtypeType,
        BackendModeType,
    )
except ImportError:
    raise ImportError("Please install mlstm_kernels package to use mLSTM block.")


import torch
from torch import nn
from typing import Literal
from .components import MultiHeadLayerNorm, RMSNorm, soft_cap
from .utils import round_up_to_next_multiple_of
from .generate import generate_tokens, get_sampling_fn

mLSTMLayerStateType = tuple[torch.Tensor, torch.Tensor, torch.Tensor]
mLSTMStateType = dict[int, mLSTMLayerStateType]

WeightModeType = Literal["single", "fused"]


@dataclass
class xLSTMLargeConfig:
    embedding_dim: int
    """Embedding dimension of the model."""
    num_heads: int
    """Number of heads."""
    num_blocks: int
    """Number of blocks."""
    vocab_size: int
    """Vocabulary size."""
    use_bias: bool = False
    """Whether to use bias in linear layers."""
    norm_eps: float = 1e-6
    """Epsilon value for numerical stability in the normalization layers."""
    norm_reduction_force_float32: bool = True
    """Whether to force float32 reductions in the normalization layers."""
    add_out_norm: bool = True
    """Whether to add a normalization layer after the block stack."""

    # mlstm layer
    qk_dim_factor: float = 0.5
    """The factor to determine the dimension of the query and key tensors."""
    v_dim_factor: float = 1.0
    """The factor to determine the dimension of the value tensor."""

    # mlstm backend
    chunkwise_kernel: ChunkwiseKernelType = "chunkwise--triton_limit_chunk"
    """Kernel to use for chunkwise parallel processing of the sequence.
    Also supports fully parallel (i.e. quadratic) backends for comparison.
    E.g. 'parallel--native_autograd'.
    """
    sequence_kernel: SequenceKernelType = "native_sequence__triton_step"
    """The sequence kernel to use for processing sequneces step-by-step.
    Used only for parts of the prefill sequence in inference mode.
    """
    step_kernel: StepKernelType = "triton"
    """The step kernel to use for processing a single step.
    Used for generation in inference mode.
    """
    mode: BackendModeType = "train"
    """The mode of operation for the backend. Determines how the `forward` method behaves.
    Available modes are 'train', 'train_with_padding', 'inference'.
    'inference' works with arbitrary sequence lengths, and does not support training. 
    It calls a sequence of different kernels to process the sequence.
    'train_with_padding' pads the input to multiples of `chunk_size`.
    """
    chunk_size: int = 64
    """The chunk size of the chunkwise kernel.
    If `mode` is 'train_with_padding', the inputs are padded to multiples of this size.
    """
    return_last_states: bool = False
    """Whether to return the last states of the sequence in training mode.
    Inference mode always returns the last states.
    """
    autocast_kernel_dtype: DtypeType = "bfloat16"
    """The dtype to use for autocast behavior in the kernel.
    If autocast is enabled all inputs are cast to this dtype before the kernel is called.
    """
    eps: float = 1e-6
    """Epsilon value for numerical stability in the kernel."""
    inference_state_dtype: DtypeType = "float32"
    """The dtype to use for the state tensors in inference mode."""
    # feedforward
    ffn_proj_factor: float = 2.6667
    """The factor to determine the dimension of the intermediate projection in the feedforward layer."""
    ffn_round_up_to_multiple_of: int = 64
    """Round the intermediate projection dimension to the next multiple of this value."""
    
    # capping
    gate_soft_cap: float = 15.0
    """Soft cap value for the gates."""
    output_logit_soft_cap: float = 30.0
    """Soft cap value for the output logits."""

    weight_mode: WeightModeType = "single"
    """The weight mode to use for the mLSTM layer.
    Mode 'single' uses separate weights for the query, key, value, and gates.
    Mode 'fused' uses a single weight matrix for the query, key, value, and gates.
    'fused' is benefitial in inference settings.
    """


class xLSTMLarge(nn.Module):
    config_class = xLSTMLargeConfig

    def __init__(self, config: xLSTMLargeConfig):
        super().__init__()
        self.config = config
        self.embedding = nn.Embedding(config.vocab_size, config.embedding_dim)

        self.backbone = xLSTMLargeBlockStack(config)

        self.lm_head = nn.Linear(
            in_features=config.embedding_dim, out_features=config.vocab_size, bias=False
        )

    def forward(
        self, x: torch.Tensor, state: mLSTMStateType | None = None
    ) -> torch.Tensor | tuple[torch.Tensor, mLSTMStateType]:
        """Forward pass of the model.
        
        Args:
            x: Input tensor of shape [B, S].
            state: State dictionary of the model. 
                   If None, the state is initialized, the model starts from an empty initial state.
        
        Returns:
            logits: Logits tensor of shape [B, S, V].
            Tuple of logits and state: State dictionary of the model, if `return_last_states` is True.
        """

        assert x.ndim == 2, f"Input must have shape [B, S], got {x.shape}"
        B, S = x.shape

        x = self.embedding(x)

        x, state = self.backbone(x, state)

        logits = self.lm_head(x)
        logits_capped = soft_cap(logits, self.config.output_logit_soft_cap)
        if self.config.return_last_states:
            return logits_capped, state
        else:
            return logits_capped

    def generate(
        self,
        prefill_tokens: torch.Tensor,
        max_length: int,
        sampling_type: str = "greedy",
        state: mLSTMStateType | None = None,
    ) -> tuple[torch.Tensor, mLSTMStateType]:
        """Generate tokens from the model.

        Args:
            prefill_tokens: Tensor of shape [B, S] with the prefill tokens.
            max_length: Maximum length of the generated sequence.
            sampling_type: Sampling type to use, e.g. 'greedy'.
            state: State dictionary of the model. 
                   If None, the state is initialized, the model starts from an empty initial state.
        
        Returns:
            tokens: Generated tokens tensor of shape [B, S].
            state: State dictionary of the model after the last generation step.
        """
        sampling_fn = get_sampling_fn(sampling_type)
        tokens, state = generate_tokens(
            llm_forward=self.forward,
            prefill_tokens=prefill_tokens,
            max_length=max_length,
            token_sample_fn=sampling_fn,
            state=state,
            device=str(self.embedding.weight.device),
        )
        return tokens, state


class xLSTMLargeBlockStack(nn.Module):
    config_class = xLSTMLargeConfig

    def __init__(self, config: xLSTMLargeConfig):
        super().__init__()
        self.config = config

        self.blocks = nn.ModuleList(
            [mLSTMBlock(config) for _ in range(config.num_blocks)]
        )

        if self.config.add_out_norm:
            self.out_norm = RMSNorm(
                num_features=config.embedding_dim,
                eps=config.norm_eps,
                use_weight=True,
                use_bias=config.use_bias,
                force_float32_reductions=config.norm_reduction_force_float32,
            )
        else:
            self.out_norm = nn.Identity()

    def forward(
        self, x: torch.Tensor, state: mLSTMStateType | None = None
    ) -> tuple[torch.Tensor, mLSTMStateType]:
        if state is None:
            state = {i: None for i in range(len(self.blocks))}

        for i, block in enumerate(self.blocks):
            block_state = state[i]
            x, block_state_new = block(x, block_state)

            if block_state is None:
                state[i] = block_state_new
            else:
                # layer state is a tuple of three tensors: c, n, m
                # we update the state in place in order to avoid creating new tensors
                for state_idx in range(len(block_state)):
                    state[i][state_idx].copy_(block_state_new[state_idx])

        x = self.out_norm(x)

        return x, state


class FeedForward(nn.Module):
    def __init__(self, config: xLSTMLargeConfig):
        super().__init__()
        self.config = config

        self.up_proj_dim = round_up_to_next_multiple_of(
            config.embedding_dim * config.ffn_proj_factor,
            config.ffn_round_up_to_multiple_of,
        )

        if self.config.weight_mode == "single":
            self.proj_up_gate = nn.Linear(
                in_features=config.embedding_dim,
                out_features=self.up_proj_dim,
                bias=self.config.use_bias,
            )
            self.proj_up = nn.Linear(
                in_features=config.embedding_dim,
                out_features=self.up_proj_dim,
                bias=self.config.use_bias,
            )
        elif self.config.weight_mode == "fused":
            self.proj_up_gate_z = nn.Linear(
                in_features=config.embedding_dim,
                out_features=2 * self.up_proj_dim,
                bias=self.config.use_bias,
            )

        self.proj_down = nn.Linear(
            in_features=self.up_proj_dim,
            out_features=config.embedding_dim,
            bias=self.config.use_bias,
        )

        self.act_fn = nn.SiLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.config.weight_mode == "single":
            x = self.act_fn(self.proj_up_gate(x)) * self.proj_up(x)
        elif self.config.weight_mode == "fused":
            x = self.proj_up_gate_z(x)
            gate, z = torch.tensor_split(x, (self.up_proj_dim,), dim=-1)
            x = self.act_fn(gate) * z

        y = self.proj_down(x)
        return y


@dataclass
class mLSTMLayerConfig:
    embedding_dim: int
    """Embedding dimension of the model."""
    num_heads: int
    """Number of heads."""
    use_bias: bool = False
    """Whether to use bias in linear layers."""
    norm_eps: float = 1e-6
    """Epsilon value for numerical stability in the normalization layers."""
    norm_reduction_force_float32: bool = True
    """Whether to force float32 reductions in the normalization layers."""

    qk_dim_factor: float = 0.5
    """The factor to determine the dimension of the query and key tensors."""
    v_dim_factor: float = 1.0
    """The factor to determine the dimension of the value tensor."""
    gate_soft_cap: float = 15.0
    """Soft cap value for the gates."""

    mlstm_backend: mLSTMBackendConfig = field(default_factory=mLSTMBackendConfig)
    """Configuration of the mLSTM backend."""

    weight_mode: WeightModeType = "single"
    """The weight mode to use for the mLSTM layer.
    Mode 'single' uses separate weights for the query, key, value, and gates.
    Mode 'fused' uses a single weight matrix for the query, key, value, and output gates.
    """


class mLSTMLayer(nn.Module):
    def __init__(self, config: mLSTMLayerConfig):
        super().__init__()
        self.config = config

        self.v_dim = int(config.embedding_dim * config.v_dim_factor)
        self.qk_dim = int(config.embedding_dim * config.qk_dim_factor)

        if self.config.weight_mode == "single":
            self.q = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=self.qk_dim,
                bias=self.config.use_bias,
            )
            self.k = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=self.qk_dim,
                bias=self.config.use_bias,
            )
            self.v = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=self.v_dim,
                bias=self.config.use_bias,
            )

            self.ogate_preact = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=self.v_dim,
                bias=self.config.use_bias,
            )
            self.igate_preact = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=self.config.num_heads,
                bias=True,
            )
            self.fgate_preact = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=self.config.num_heads,
                bias=True,
            )
        elif self.config.weight_mode == "fused":
            self.qkv_opreact = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=2 * self.qk_dim + 2 * self.v_dim,
                bias=self.config.use_bias,
            )
            self.ifgate_preact = nn.Linear(
                in_features=self.config.embedding_dim,
                out_features=2 * self.config.num_heads,
                bias=True,
            )

        self.ogate_act_fn = nn.Sigmoid()
        self.mlstm_backend = mLSTMBackend(config=self.config.mlstm_backend)

        self.multihead_norm = MultiHeadLayerNorm(
            num_heads=self.config.num_heads,
            head_dim=self.v_dim // self.config.num_heads,
            eps=self.config.norm_eps,
            use_weight=True,
            use_bias=self.config.use_bias,
            force_float32_reductions=self.config.norm_reduction_force_float32,
        )
        self.out_proj = nn.Linear(
            in_features=self.v_dim,
            out_features=self.config.embedding_dim,
            bias=self.config.use_bias,
        )

    def forward(
        self, x: torch.Tensor, state: mLSTMLayerStateType | None = None
    ) -> tuple[torch.Tensor, mLSTMLayerStateType | None]:
        assert x.ndim == 3, f"Input must have shape [B, S, D], got {x.shape}"
        B, S, _ = x.shape
        if self.config.weight_mode == "single":
            q = self.q(x)
            k = self.k(x)
            v = self.v(x)
            o_preact = self.ogate_preact(x)
            i_preact = soft_cap(
                self.igate_preact(x), cap_value=self.config.gate_soft_cap
            )
            f_preact = soft_cap(
                self.fgate_preact(x), cap_value=self.config.gate_soft_cap
            )

        elif self.config.weight_mode == "fused":
            qkv_opreact = self.qkv_opreact(x)
            q, k, v, o_preact = torch.tensor_split(
                qkv_opreact,
                (
                    self.qk_dim,
                    2 * self.qk_dim,
                    2 * self.qk_dim + self.v_dim,
                ),
                dim=-1,
            )

            if_preact = soft_cap(
                self.ifgate_preact(x), cap_value=self.config.gate_soft_cap
            )
            i_preact, f_preact = torch.tensor_split(
                if_preact, (self.config.num_heads,), dim=-1
            )

        q = q.reshape(B, S, self.config.num_heads, -1).transpose(1, 2)
        k = k.reshape(B, S, self.config.num_heads, -1).transpose(1, 2)
        v = v.reshape(B, S, self.config.num_heads, -1).transpose(1, 2)
        i_preact = i_preact.transpose(1, 2)
        f_preact = f_preact.transpose(1, 2)
        if state is None:
            c_initial, n_initial, m_initial = None, None, None
        else:
            c_initial, n_initial, m_initial = state

        h, state = self.mlstm_backend(
            q=q,
            k=k,
            v=v,
            i=i_preact,
            f=f_preact,
            c_initial=c_initial,
            n_initial=n_initial,
            m_initial=m_initial,
        )
        expected_h_shape = (
            B,
            self.config.num_heads,
            S,
            self.v_dim // self.config.num_heads,
        )
        assert (
            h.shape == expected_h_shape
        ), f"Got {h.shape}, expected {expected_h_shape}"

        h = h.transpose(1, 2)
        h_norm = self.multihead_norm(h)
        h_norm = h_norm.reshape(B, S, -1)

        h_out = self.ogate_act_fn(o_preact) * h_norm

        y = self.out_proj(h_out)
        return y, state


class mLSTMBlock(nn.Module):
    def __init__(self, config: xLSTMLargeConfig):
        super().__init__()
        self.config = config
        self.norm_mlstm = RMSNorm(
            num_features=config.embedding_dim,
            eps=config.norm_eps,
            use_weight=True,
            use_bias=config.use_bias,
            force_float32_reductions=config.norm_reduction_force_float32,
        )
        self.mlstm_layer = mLSTMLayer(
            mLSTMLayerConfig(
                embedding_dim=config.embedding_dim,
                num_heads=config.num_heads,
                use_bias=config.use_bias,
                norm_eps=config.norm_eps,
                norm_reduction_force_float32=config.norm_reduction_force_float32,
                qk_dim_factor=config.qk_dim_factor,
                v_dim_factor=config.v_dim_factor,
                gate_soft_cap=config.gate_soft_cap,
                weight_mode=config.weight_mode,
                mlstm_backend=mLSTMBackendConfig(
                    chunkwise_kernel=config.chunkwise_kernel,
                    sequence_kernel=config.sequence_kernel,
                    step_kernel=config.step_kernel,
                    mode=config.mode,
                    chunk_size=config.chunk_size,
                    return_last_states=config.return_last_states,
                    autocast_kernel_dtype=config.autocast_kernel_dtype,
                    eps=config.eps,
                    inference_state_dtype=config.inference_state_dtype,
                ),
            )
        )
        self.norm_ffn = RMSNorm(
            num_features=config.embedding_dim,
            eps=config.norm_eps,
            use_weight=True,
            use_bias=config.use_bias,
            force_float32_reductions=config.norm_reduction_force_float32,
        )
        self.ffn = FeedForward(config)

    def forward(
        self, x: torch.Tensor, state: mLSTMStateType | None = None
    ) -> tuple[torch.Tensor, mLSTMStateType]:
        x_mlstm = self.norm_mlstm(x)
        x_mlstm, state = self.mlstm_layer(x_mlstm, state)
        x = x + x_mlstm

        x_ffn = self.norm_ffn(x)
        x_ffn = self.ffn(x_ffn)
        x = x + x_ffn

        return x, state



# Copyright (c) NXAI GmbH and its affiliates 2023
# Korbinian Poeppel
import logging
import os
from dataclasses import dataclass, field
from typing import Optional, Literal
from pathlib import Path

import torch

from torch.autograd.function import once_differentiable
from .src.cuda_init import load
from .src.vanilla import (
    slstm_forward,
    slstm_forward_step,
    slstm_pointwise_function_registry,
)
from ...components.util import conditional_decorator, round_to_multiple, ParameterProxy
from ...components.init import bias_linspace_init_

from dataclasses import dataclass, field
from typing import Callable, Literal, Optional, Union, Sequence
import logging
import torch
import torch.nn as nn
from math import sqrt

LOGGER = logging.getLogger(__name__)

DTYPE_DICT = {
    "bfloat16": torch.bfloat16,
    "float16": torch.float16,
    "float32": torch.float32,
}
DTYPES = Literal["bfloat16", "float16", "float32"]

# this is needed to locate the cuda source file path
curdir = Path(os.path.split(os.path.os.path.abspath(__file__))[0])

# maps the rnn function to the following values
#   states = 2 or 4

rnn_function_registry = {
    # standard variants, all connect
    "lstm": {
        "states": 2,
    },
    "slstm": {
        "states": 4,
    },
}

_python_dtype_to_cuda_dtype = {
    "float32": "float",
    "float": "float",
    "float16": "__half",
    "bfloat16": "__nv_bfloat16",
}


@dataclass
class sLSTMCellConfig:
    hidden_size: int = -1
    num_heads: int = (
        4  # this must divide the hidden size, is not yet supported by all versions in this directory
    )
    num_states: int = 4  # this is for the sLSTM, a standard LSTM  has 2
    backend: Literal["vanilla", "cuda"] = "cuda"
    # the type of function a cell computes
    function: str = "slstm"
    bias_init: Literal["powerlaw_blockdependent", "small_init", "standard"] = (
        "powerlaw_blockdependent"
    )
    recurrent_weight_init: Literal["zeros", "standard"] = "zeros"

    _block_idx: int = (
        0  # index in the block sequence for a residual stacked model, needed for forget gate init
    )
    _num_blocks: int = (
        1  # how many blocks there are in the residually stacked model, needed for forget gate init
    )

    num_gates: int = 4
    # this option cuts of the gradient for recurrent connection, i.e. no exploding gradient if False
    gradient_recurrent_cut: bool = False
    # this option clips the gradient values for recurrent connections at dy
    gradient_recurrent_clipval: Optional[float] = None
    # this option clips the y value
    forward_clipval: Optional[float] = None

    # this can be ignored internally, but may be used to optimize kernels
    batch_size: int = 8
    # input shape from external, definitions see output_shape
    input_shape: Literal["BSGNH", "SBGNH"] = "BSGNH"
    # internal input shape, may be redefined by backend
    internal_input_shape: Literal["SBNGH", "SBGNH", "SBNHG"] = "SBNGH"
    # B = batch, S sequence dim, N num heads,
    # H head dim or hidden dim (= fused dim [num_heads, head_dim]) without num_heads
    output_shape: Literal[
        "BNSH",
        "SBH",
        "BSH",
        "SBNH",
    ] = "BNSH"

    # additional compiler constants
    constants: dict = field(default_factory=dict)
    # dtypes
    dtype: DTYPES = "bfloat16"
    dtype_b: Optional[DTYPES] = "float32"  # biases
    dtype_r: Optional[DTYPES] = None  # recurrent matrix
    dtype_w: Optional[DTYPES] = None  # inputs / w matrix
    dtype_g: Optional[DTYPES] = None  # gates
    dtype_s: Optional[DTYPES] = None  # states
    dtype_a: Optional[DTYPES] = None  # internal accumulation

    # mixed precision
    enable_automatic_mixed_precision: bool = True
    initial_val: Union[float, Sequence[float]] = 0.0

    @property
    def head_dim(self):
        return self.hidden_size // self.num_heads

    @property
    def input_dim(self):
        return 4 * self.hidden_size

    @property
    def torch_dtype(self) -> torch.dtype:
        return DTYPE_DICT[self.dtype]

    @property
    def torch_dtype_b(self) -> torch.dtype:
        return DTYPE_DICT[self.dtype_b]

    @property
    def torch_dtype_r(self) -> torch.dtype:
        return DTYPE_DICT[self.dtype_r]

    @property
    def torch_dtype_w(self) -> torch.dtype:
        return DTYPE_DICT[self.dtype_w]

    @property
    def torch_dtype_s(self) -> torch.dtype:
        return DTYPE_DICT[self.dtype_s]

    def __post_init__(self):
        if self.num_heads <= 0:
            self.num_heads = 1
        if self.dtype_b is None:
            self.dtype_b = self.dtype
        if self.dtype_a is None:
            self.dtype_a = self.dtype_b
        if self.dtype_r is None:
            self.dtype_r = self.dtype
        if self.dtype_w is None:
            self.dtype_w = self.dtype
        if self.dtype_s is None:
            self.dtype_s = self.dtype_w
        if self.dtype_g is None:
            self.dtype_g = self.dtype_r

        assert (
            self.function in rnn_function_registry
        ), f"RNN function {self.function} not in registry"
        self.num_states = rnn_function_registry[self.function]["states"]
        if "initial_val" in rnn_function_registry[self.function]:
            self.initial_val = rnn_function_registry[self.function]["initial_val"]

    @property
    def defines(self):
        return (
            [
                f"-DSLSTM_HIDDEN_SIZE={self.hidden_size}",
                f"-DSLSTM_BATCH_SIZE={self.batch_size}",
                f"-DSLSTM_NUM_HEADS={self.num_heads}",
                f"-DSLSTM_NUM_STATES={self.num_states}",
                f"-DSLSTM_DTYPE_B={_python_dtype_to_cuda_dtype[self.dtype_b]}",
                f"-DSLSTM_DTYPE_R={_python_dtype_to_cuda_dtype[self.dtype_r]}",
                f"-DSLSTM_DTYPE_W={_python_dtype_to_cuda_dtype[self.dtype_w]}",
                f"-DSLSTM_DTYPE_G={_python_dtype_to_cuda_dtype[self.dtype_g]}",
                f"-DSLSTM_DTYPE_S={_python_dtype_to_cuda_dtype[self.dtype_s]}",
                f"-DSLSTM_DTYPE_A={_python_dtype_to_cuda_dtype[self.dtype_a]}",
                f"-DSLSTM_NUM_GATES={4}",
                f"-DSLSTM_SIMPLE_AGG={'true'}",
            ]
            + (
                [
                    f"-DSLSTM_GRADIENT_RECURRENT_CLIPVAL_VALID=true",
                    f"-DSLSTM_GRADIENT_RECURRENT_CLIPVAL={self.gradient_recurrent_clipval}",
                ]
                if self.gradient_recurrent_clipval is not None
                else [
                    f"-DSLSTM_GRADIENT_RECURRENT_CLIPVAL_VALID=false",
                    f"-DSLSTM_GRADIENT_RECURRENT_CLIPVAL=0.0",
                ]
            )
            + (
                [
                    f"-DSLSTM_FORWARD_CLIPVAL_VALID=true",
                    f"-DSLSTM_FORWARD_CLIPVAL={self.gradient_recurrent_clipval}",
                ]
                if self.gradient_recurrent_clipval is not None
                else [
                    f"-DSLSTM_FORWARD_CLIPVAL_VALID=false",
                    f"-DSLSTM_FORWARD_CLIPVAL=0.0",
                ]
            )
        )


class sLSTMCellBase(nn.Module):
    config_class = sLSTMCellConfig

    def __init__(self, config: sLSTMCellConfig):
        super().__init__()
        self.config = config
        LOGGER.debug("Init module")

        head_dim = self.config.hidden_size // self.config.num_heads
        dtype_r = (
            self.config.torch_dtype_r
            if not self.config.enable_automatic_mixed_precision
            else None
        )
        dtype_b = (
            self.config.torch_dtype_b
            if not self.config.enable_automatic_mixed_precision
            else None
        )

        self._recurrent_kernel_ = nn.Parameter(
            torch.empty(
                self.config.num_heads,
                head_dim,
                self.config.num_gates,
                head_dim,
                dtype=dtype_r,
            )
        )
        self.recurrent_kernel = ParameterProxy(
            self,
            "_recurrent_kernel",
            self._recurrent_kernel_int2ext,
            self._recurrent_kernel_ext2int,
        )
        self._recurrent_kernel_ = nn.Parameter(
            self._recurrent_kernel_ext2int(self._recurrent_kernel_.data)
        )

        self._bias_ = nn.Parameter(
            torch.empty(
                self.config.num_heads, self.config.num_gates, head_dim, dtype=dtype_b
            )
        )
        self.bias = ParameterProxy(
            self, "_bias", self._bias_int2ext, self._bias_ext2int
        )
        self._bias_ = nn.Parameter(self._bias_ext2int(self._bias_.data))

        self.reset_parameters()

        if self.config.hidden_size % self.config.num_heads != 0:
            raise ValueError(
                f"Hidden Size {self.config.hidden_size} must be divisible by head num {self.config.num_heads}"
            )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(function={self.config.function}, "
            f"hidden_size={self.config.hidden_size}, num_heads={self.config.num_heads})"
        )

    @property
    def _recurrent_kernel(self):
        return self._recurrent_kernel_

    @property
    def _bias(self):
        return self._bias_

    def _recurrent_kernel_ext2int(
        self, recurrent_kernel_ext: torch.Tensor
    ) -> torch.Tensor:
        return recurrent_kernel_ext

    def _bias_ext2int(self, bias_ext: torch.Tensor) -> torch.Tensor:
        return bias_ext

    def _recurrent_kernel_int2ext(
        self, recurrent_kernel_int: torch.Tensor
    ) -> torch.Tensor:
        return recurrent_kernel_int

    def _bias_int2ext(self, bias_int: torch.Tensor) -> torch.Tensor:
        return bias_int

    def parameters_to_dtype(self):
        pars = [name for name, _ in self.named_parameters()]
        for name in pars:
            par = getattr(self, name)
            if "recurrent" in name:
                setattr(
                    self,
                    name,
                    torch.nn.Parameter(
                        par.to(dtype=self.config.dtype_r),
                        requires_grad=par.requires_grad,
                    ),
                )
            if "bias" in name:
                setattr(
                    self,
                    name,
                    torch.nn.Parameter(
                        par.to(dtype=self.config.dtype_b),
                        requires_grad=par.requires_grad,
                    ),
                )

    @property
    def head_dim(self):
        return self.config.hidden_size // self.config.num_heads

    def _permute_input(self, x: torch.Tensor) -> torch.Tensor:
        """
        # TODO Adapt this
        # >>> BaseRNN(BaseRNNConfig(10, 10, num_heads=2, input_shape='SBG'))._permute_input(torch.zeros((5, 2, 10))).shape
        # torch.Size([5, 2, 10])
        # >>> BaseRNN(BaseRNNConfig(10, 10, num_heads=2, input_shape='BSG'))._permute_input(torch.zeros((5, 2, 10))).shape
        # torch.Size([2, 5, 10])
        """
        if self.config.input_shape == "SBGNH":
            y = x.view(
                x.shape[0], x.shape[1], self.config.num_gates, self.config.num_heads, -1
            )
        elif self.config.input_shape == "BSGNH":
            y = x.view(
                x.shape[0], x.shape[1], self.config.num_gates, self.config.num_heads, -1
            ).permute(1, 0, 2, 3, 4)
        else:
            raise ValueError("Bad input_shape value")
        if self.config.internal_input_shape == "SBGNH":
            return y.view(y.shape[0], y.shape[1], -1)
        elif self.config.internal_input_shape == "SBNGH":
            return y.permute(0, 1, 3, 2, 4).reshape(y.shape[0], y.shape[1], -1)
        elif self.config.internal_input_shape == "SBNHG":
            return y.permute(0, 1, 3, 4, 2).reshape(y.shape[0], y.shape[1], -1)
        else:
            raise ValueError("Bad internal_input_shape value")

    def _permute_output(self, x: torch.Tensor) -> torch.Tensor:
        """
        >>> BaseRNN(BaseRNNConfig(10, 16, num_heads=2, output_shape='SBH'))._permute_output(torch.zeros((5, 2, 16))).shape
        torch.Size([5, 2, 16])
        >>> BaseRNN(BaseRNNConfig(10, 16, num_heads=2, output_shape='BNSH'))._permute_output(torch.zeros((5, 3, 16))).shape
        torch.Size([3, 2, 5, 8])
        >>> BaseRNN(BaseRNNConfig(10, 16, num_heads=2, output_shape='SBNH'))._permute_output(torch.zeros((5, 3, 16))).shape
        torch.Size([5, 3, 2, 8])
        """
        if self.config.output_shape == "SBH":
            return x
        elif self.config.output_shape == "BSH":
            return x.permute(1, 0, 2)
        elif self.config.output_shape == "BNSH":
            return x.view(
                (x.shape[0], x.shape[1], self.config.num_heads, self.config.head_dim)
            ).permute(1, 2, 0, 3)
        elif self.config.output_shape == "SBNH":
            return x.view(
                (x.shape[0], x.shape[1], self.config.num_heads, self.config.head_dim)
            )

    def reset_parameters(self):
        """Resets this layer's parameters to their initial values."""
        for h in range(self.config.num_heads):
            for i, gate in enumerate(["i", "f", "z", "o"]):
                if self.config.recurrent_weight_init == "zeros":
                    self.recurrent_kernel[h, :, i, :] = nn.init.zeros_(
                        self.recurrent_kernel[h, :, i, :]
                    )
                elif self.config.recurrent_weight_init == "standard":
                    self.recurrent_kernel[h, :, i, :] = nn.init.uniform_(
                        self.recurrent_kernel[h, :, i, :],
                        -1.0 / sqrt(self.config.hidden_size),
                        1.0 / sqrt(self.config.hidden_size),
                    )
        for h in range(self.config.num_heads):
            for i, gate in enumerate(["i", "f", "z", "o"]):
                if self.config.bias_init == "powerlaw_blockdependent":
                    if gate == "f":
                        kwargs = (
                            dict(
                                right_x=5.0,
                                range_x_neg_dir=12.0,
                                spread_lower=0.3,
                                spread_upper=3.0,
                            ),
                        )
                        ratio_0_to_1 = (
                            self.config._block_idx / (self.config._num_blocks - 1)
                            if self.config._num_blocks > 1
                            else 0.0
                        )
                        init_values = -(
                            -5.0
                            + 12.0
                            * (
                                torch.arange(self.config.head_dim)
                                / (self.config.head_dim - 1)
                            )
                            ** (0.3 + 1.3 * ratio_0_to_1)
                        )
                        with torch.no_grad():
                            self.bias[h, i, :] = init_values
                    else:
                        self.bias[h, i] = nn.init.zeros_(self.bias[h, i])
                elif self.config.bias_init == "small_init":
                    if gate == "f":
                        self.bias[h, i] = bias_linspace_init_(
                            self.bias[h, i], start=3.0, end=6.0
                        )
                    else:
                        self.bias[h, i] = nn.init.zeros_(self.bias[h, i])
                elif self.config.bias_init == "zeros":
                    self.bias[h, i] = nn.init.zeros_(self.bias[h, i])
                elif self.config.bias_init == "standard":
                    self.bias[h, i] = nn.init.uniform_(
                        self.bias[h, i],
                        -1 / sqrt(self.config.hidden_size),
                        1 / sqrt(self.config.hidden_size),
                    )

    def _check_input(self, input: torch.Tensor) -> None:
        assert self.config.hidden_size * self.config.num_gates == input.size(
            -1
        ), f"Input size mismatch: Expected input size {self.config.hidden_size * self.config.num_gates}, but got {input.size(-1)}."

    def _zero_state(self, input: torch.Tensor) -> torch.Tensor:
        """Returns a zeros state matching dtype and batch size of `input`.

        Arguments:
          input: Tensor, to specify the device and dtype of the returned tensors.

        Returns:
          zero_state: a nested structure of zero Tensors.
        """
        batch_dim = input.shape[1]
        state = torch.zeros(
            (self.config.num_states, batch_dim, self.config.hidden_size),
            dtype=input.dtype,
            device=input.device,
        )
        return state

    def _get_state(
        self, input: torch.Tensor, state: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        if state is None:
            state = self._zero_state(input)
        else:
            assert state.shape == (
                self.config.num_states,
                input.shape[1],
                self.config.hidden_size,
            )
        return state

    def _get_final_state(self, all_states: torch.Tensor) -> torch.Tensor:
        """
        All states has the structure
        [STATES, SEQUENCE, BATCH, HIDDEN]
        """
        return all_states[:, -1]

    def _is_cuda(self) -> bool:
        is_cuda = [tensor.is_cuda for tensor in list(self.parameters())]
        if any(is_cuda) and not all(is_cuda):
            raise ValueError(
                "RNN tensors should all be CUDA tensors or none should be CUDA tensors"
            )
        return any(is_cuda)

    def step(
        self, input: torch.Tensor, state: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        self._check_input(input)
        input = self._permute_input(input)
        states = self._get_state(input, state)
        all_states = self._impl_step(self.training, input, states)
        output = self._permute_output(all_states[0])
        return output, state

    def forward(self, input, state=None, lengths=None):
        self._check_input(input)
        input = self._permute_input(input)
        states = self._get_state(input, state)
        all_states = self._impl(self.training, input, states)
        state = self._get_final_state(all_states)
        output = self._permute_output(all_states[0][1:])
        if torch.is_autocast_enabled():
            return output, state
        else:
            return output.to(input.dtype), state.to(input.dtype)


class sLSTMCellCUDA(object):
    mod = {}

    @classmethod
    def instance(cls, config: sLSTMCellConfig):
        if repr(config) not in cls.mod:
            cls.mod[repr(config)] = load(
                name=config.function,
                sources=[
                    str(curdir / "src" / "cuda" / "slstm.cc"),
                    str(curdir / "src" / "cuda" / "slstm_forward.cu"),
                    str(curdir / "src" / "cuda" / "slstm_backward.cu"),
                    str(curdir / "src" / "cuda" / "slstm_backward_cut.cu"),
                    str(curdir / "src" / "cuda" / f"{config.function}_pointwise.cu"),
                    str(curdir / "src" / "util" / "blas.cu"),
                    str(curdir / "src" / "util" / "cuda_error.cu"),
                ],
                extra_cflags=[
                    f"-D{const}={constval}"
                    for const, constval in config.constants.items()
                ]
                + config.defines,
            )
        return cls.mod[repr(config)]


def sLSTMCellFuncGenerator(training, config: sLSTMCellConfig):
    slstm_cuda = sLSTMCellCUDA.instance(config=config)
    slstm_mod = slstm_cuda.sLSTMFunc(
        training, config.batch_size, config.hidden_size, config.num_heads
    )

    class sLSTMCellFunction(torch.autograd.Function):
        @staticmethod
        @conditional_decorator(
            config.enable_automatic_mixed_precision, torch.cuda.amp.custom_fwd
        )
        def forward(ctx, training, *inputs):
            dtypes = (
                inputs[0].dtype,
                inputs[1].dtype,
                inputs[2].dtype,
                inputs[3].dtype,
            )
            if config.enable_automatic_mixed_precision:
                inputs = (
                    inputs[0].to(dtype=config.torch_dtype_w),
                    inputs[1].to(dtype=config.torch_dtype_s),
                    inputs[2].to(dtype=config.torch_dtype_r),
                    inputs[3].to(dtype=config.torch_dtype_b),
                )
            states, cache_g_r, cache_g_i = slstm_mod.forward(training, *inputs)

            ctx.save_for_backward(*inputs[2:], states, cache_g_r, cache_g_i)
            ctx.training = training
            return states

        @staticmethod
        @once_differentiable
        @conditional_decorator(
            config.enable_automatic_mixed_precision, torch.cuda.amp.custom_bwd
        )
        def backward(ctx, grad_s):
            if not ctx.training:
                raise RuntimeError(
                    "sLSTMCell backward can only be called in training mode"
                )
            saved = [*ctx.saved_tensors]
            saved[0] = saved[0].permute(0, 2, 1).contiguous()  # transpose R
            if config.gradient_recurrent_cut:
                grads = slstm_mod.backward_cut(*saved, grad_s.contiguous())
            else:
                grads = slstm_mod.backward(*saved, grad_s.contiguous())
            with torch.no_grad():
                S, B, H = grads[0].shape
            return (None, *grads)

    return sLSTMCellFunction


class sLSTMCell_vanilla(sLSTMCellBase):
    config_class = sLSTMCellConfig

    def __init__(self, config: sLSTMCellConfig, skip_backend_init=False):
        super().__init__(config)
        # load pointwise function
        self.pointwise = slstm_pointwise_function_registry[self.config.function]

        self.config.internal_input_shape = "SBGNH"

    def _recurrent_kernel_ext2int(
        self, recurrent_kernel_ext: torch.Tensor
    ) -> torch.Tensor:
        return (
            recurrent_kernel_ext.reshape(
                self.config.num_heads,
                self.config.head_dim,
                self.config.num_gates,
                self.config.head_dim,
            )
            .permute(0, 2, 3, 1)
            .reshape(
                self.config.num_heads,
                self.config.num_gates * self.config.head_dim,
                self.config.head_dim,
            )
        )

    def _recurrent_kernel_int2ext(
        self, recurrent_kernel_int: torch.Tensor
    ) -> torch.Tensor:
        """
        >>> (); mod = sLSTMCell_vanilla(sLSTMCellConfig(hidden_size=64, num_heads=2), skip_backend_init=True); () # doctest:+ELLIPSIS
        (...)
        >>> torch.allclose(mod._recurrent_kernel_ext2int(mod._recurrent_kernel_int2ext(mod._recurrent_kernel)), mod._recurrent_kernel)
        True
        """
        return recurrent_kernel_int.reshape(
            self.config.num_heads,
            self.config.num_gates,
            self.config.head_dim,
            self.config.head_dim,
        ).permute(0, 3, 1, 2)

    def _bias_ext2int(self, bias_ext: torch.Tensor) -> torch.Tensor:
        return (
            bias_ext.reshape(
                self.config.num_heads, self.config.num_gates, self.config.head_dim
            )
            .permute(1, 0, 2)
            .reshape(-1)
        )

    def _bias_int2ext(self, bias_int: torch.Tensor) -> torch.Tensor:
        """
        >>> (); mod = sLSTMCell_vanilla(sLSTMCellConfig(hidden_size=64, num_heads=2), skip_backend_init=True); () # doctest:+ELLIPSIS
        (...)
        >>> torch.allclose(mod._bias_ext2int(mod._bias_int2ext(mod._bias)), mod._bias)
        True
        """

        return bias_int.reshape(
            self.config.num_gates, self.config.num_heads, self.config.head_dim
        ).permute(1, 0, 2)

    def _impl(
        self, training: bool, input: torch.Tensor, state: torch.Tensor
    ) -> torch.Tensor:
        return slstm_forward(
            input,
            state,
            self._recurrent_kernel,
            self._bias,
            self.pointwise,
            constants=self.config.constants,
        )[0]

    def _impl_step(
        self, training: bool, input: torch.Tensor, state: torch.Tensor
    ) -> torch.Tensor:
        return slstm_forward_step(
            input,
            state,
            self._recurrent_kernel,
            self._bias,
            self.pointwise,
            constants=self.config.constants,
        )[0]


class sLSTMCell_cuda(sLSTMCellBase):
    config_class = sLSTMCellConfig

    def __init__(self, config: sLSTMCellConfig, skip_backend_init: bool = False):
        """
        skip device init is intended for converting models between hardware configurations / backends
        i.e. to store a model first and later convert it to a different backend form
        """
        super().__init__(config)
        self.internal_input_shape = "SBNGH"
        if not skip_backend_init:
            self.func = sLSTMCellFuncGenerator(self.training, config)

    def _recurrent_kernel_ext2int(
        self, recurrent_kernel_ext: torch.Tensor
    ) -> torch.Tensor:
        return recurrent_kernel_ext.reshape(
            self.config.num_heads,
            self.config.head_dim,
            self.config.num_gates,
            self.config.head_dim,
        ).reshape(
            self.config.num_heads,
            self.config.head_dim,
            self.config.num_gates * self.config.head_dim,
        )

    def _recurrent_kernel_int2ext(
        self, recurrent_kernel_int: torch.tensor
    ) -> torch.Tensor:
        """
        >>> (); mod = sLSTMCell_cuda(
        ...     sLSTMCellConfig(hidden_size=64, num_heads=2), skip_backend_init=True); () # doctest:+ELLIPSIS
        (...)
        >>> torch.allclose(mod._recurrent_kernel_ext2int(mod._recurrent_kernel_int2ext(mod._recurrent_kernel)), mod._recurrent_kernel)
        True
        """

        return recurrent_kernel_int.reshape(
            self.config.num_heads,
            self.config.head_dim,
            self.config.num_gates,
            self.config.head_dim,
        )

    def _bias_ext2int(self, bias_ext: torch.Tensor) -> torch.Tensor:
        return (
            bias_ext.reshape(
                self.config.num_heads, self.config.num_gates, self.config.head_dim
            )
            .permute(0, 1, 2)
            .reshape(-1)
        )

    def _bias_int2ext(self, bias_int: torch.Tensor) -> torch.Tensor:
        """
        >>> (); mod = sLSTMCell_cuda(
        ...     sLSTMCellConfig(hidden_size=64, num_heads=2), skip_backend_init=True); () # doctest:+ELLIPSIS
        (...)
        >>> torch.allclose(mod._bias_ext2int(mod._bias_int2ext(mod._bias)), mod._bias)
        True
        """

        return bias_int.reshape(
            self.config.num_heads, self.config.num_gates, self.config.head_dim
        )

    def _impl_step(
        self,
        training: bool,
        input: torch.Tensor,
        state: torch.Tensor,
    ) -> torch.Tensor:
        return self.func.apply(
            training,
            input.contiguous(),
            state.contiguous(),
            self._recurrent_kernel.contiguous(),
            self._bias.contiguous(),
        )

    def _impl(
        self,
        training: bool,
        input: torch.Tensor,
        state: torch.Tensor,
    ) -> torch.Tensor:
        return self.func.apply(
            training,
            input.contiguous(),
            state.contiguous(),
            self._recurrent_kernel.contiguous(),
            self._bias.contiguous(),
        )


class sLSTMCell(sLSTMCellBase):
    config_class = sLSTMCellConfig

    def __new__(cls, config: sLSTMCellConfig, skip_backend_init: bool = False):
        if config.backend == "cuda":
            return sLSTMCell_cuda(config, skip_backend_init=skip_backend_init)
        elif config.backend == "vanilla":
            return sLSTMCell_vanilla(config)
        else:
            raise RuntimeError(
                f'sLSTMCell unknown backend {config.backend}, choose from ["cuda", "vanilla"]'
            )



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck, Korbinian Pöppel
import math
from typing import Optional

import torch


def parallel_stabilized_simple(
    queries: torch.Tensor,
    keys: torch.Tensor,
    values: torch.Tensor,
    igate_preact: torch.Tensor,
    fgate_preact: torch.Tensor,
    lower_triangular_matrix: torch.Tensor = None,
    stabilize_rowwise: bool = True,
    eps: float = 1e-6,
    **kwargs,
) -> torch.Tensor:
    """This is the mLSTM cell in parallel form.
    This version is stabilized. We control the range of exp() arguments by
    ensuring that they are always smaller than 0.0 by subtracting the maximum.

    Args:
        queries (torch.Tensor): (B, NH, S, DH)
        keys (torch.Tensor): (B, NH, S, DH)
        values (torch.Tensor): (B, NH, S, DH)
        igate_preact (torch.Tensor): (B, NH, S, 1)
        fgate_preact (torch.Tensor): (B, NH, S, 1)
        lower_triangular_matrix (torch.Tensor, optional): (S,S). Defaults to None.
        stabilize_rowwise (bool, optional): Wether to stabilize the combination matrix C rowwise (take maximum per row).
            Alternative: Subtract the maximum over all rows. Defaults to True.

    Returns:
        torch.Tensor: (B, NH, S, DH), h_tilde_state
    """

    B, NH, S, DH = queries.shape
    _dtype, _device = queries.dtype, queries.device

    # forget gate matrix
    log_fgates = torch.nn.functional.logsigmoid(fgate_preact)  # (B, NH, S, 1)
    if lower_triangular_matrix is None or S < lower_triangular_matrix.size(-1):
        ltr = torch.tril(torch.ones((S, S), dtype=torch.bool, device=_device))
    else:
        ltr = lower_triangular_matrix
    assert ltr.dtype == torch.bool, f"lower_triangular_matrix must be of dtype bool, got {ltr.dtype}"

    log_fgates_cumsum = torch.cat(
        [
            torch.zeros((B, NH, 1, 1), dtype=_dtype, device=_device),
            torch.cumsum(log_fgates, dim=-2),
        ],
        dim=-2,
    )  # (B, NH, S+1, 1)
    # for each batch/head this is a matrix of shape (S+1, S+1) containing the cumsum of the log forget gate values
    # in the second dimension (colum dimension). Each row has the same is a copy of the first row.
    # First entry of each row is zero.
    rep_log_fgates_cumsum = log_fgates_cumsum.repeat(1, 1, 1, S + 1)  # (B, NH, S+1, S+1)
    # Now in each row cut off / subtract the forgetgate values of the later timesteps
    # where col j > row i
    _log_fg_matrix = rep_log_fgates_cumsum - rep_log_fgates_cumsum.transpose(-2, -1)  # (B, NH, S+1, S+1)
    # Causal masking & selection of the correct submatrix, such that forgetgate at timestep t is not applied
    # to the input at timestep t
    log_fg_matrix = torch.where(ltr, _log_fg_matrix[:, :, 1:, 1:], -float("inf"))  # (B, NH, S, S)

    # gate decay matrix D (combination of forget gate and input gate)
    log_D_matrix = log_fg_matrix + igate_preact.transpose(-2, -1)  # (B, NH, S, S)
    # D matrix stabilization
    if stabilize_rowwise:
        max_log_D, _ = torch.max(log_D_matrix, dim=-1, keepdim=True)  # (B, NH, S, 1)
    else:
        max_log_D = torch.max(log_D_matrix.view(B, NH, -1), dim=-1, keepdim=True)[0].unsqueeze(-1)
        # (B, NH, 1, 1)
    log_D_matrix_stabilized = log_D_matrix - max_log_D  # (B, NH, S, S)
    D_matrix = torch.exp(log_D_matrix_stabilized)  # (B, NH, S, S)

    keys_scaled = keys / math.sqrt(DH)

    # combination matrix C
    qk_matrix = queries @ keys_scaled.transpose(-2, -1)  # (B, NH, S, S)
    C_matrix = qk_matrix * D_matrix  # (B, NH, S, S)
    normalizer = torch.maximum(C_matrix.sum(dim=-1, keepdim=True).abs(), torch.exp(-max_log_D))  # (B, NH, S, 1)
    # (B, NH, S, S)
    C_matrix_normalized = C_matrix / (normalizer + eps)

    # retrieved values
    h_tilde_state = C_matrix_normalized @ values  # (B, NH, S, DH)

    return h_tilde_state


def recurrent_step_stabilized_simple(
    c_state: torch.Tensor,
    n_state: torch.Tensor,
    m_state: torch.Tensor,
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    igate_preact: torch.Tensor,
    fgate_preact: torch.Tensor,
    eps: float = 1e-6,
    **kwargs,
) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
    """This is a single step of the mLSTM operation in recurrent form.

    Args:
        c_state (torch.Tensor): (B, NH, DH, DH)
        n_state (torch.Tensor): (B, NH, DH, 1)
        m_state (torch.Tensor): (B, NH, 1, 1)
        q (torch.Tensor): (B, NH, 1, DH)
        k (torch.Tensor): (B, NH, 1, DH)
        v (torch.Tensor): (B, NH, 1, DH)
        igate_preact (torch.Tensor): (B, NH, 1, 1)
        fgate_preact (torch.Tensor): (B, NH, 1, 1)

    Returns:
        tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
            (hidden_state [B, NH, DH], (c_state_new [B, NH, DH, DH], n_state_new [B, NH, DH, 1]], m_state_new [B, NH, 1, 1]))
    """
    B, NH, S, DH = q.shape
    # projections
    q, k, v = q.squeeze_(2).unsqueeze(-1), k.squeeze_(2).unsqueeze(-1), v.squeeze_(2).unsqueeze(-1)  # (B, NH, DH, 1)

    # gates
    log_fg_act = torch.nn.functional.logsigmoid(fgate_preact)  # (B, NH, 1, 1)

    # update rule
    m_state_new = torch.max(log_fg_act + m_state, igate_preact)  # (B, NH, 1, 1)

    fg_act = torch.exp(log_fg_act + m_state - m_state_new)  # (B, NH, 1, 1)
    ig_act = torch.exp(igate_preact - m_state_new)  # (B, NH, 1, 1)

    k_scaled = k / math.sqrt(DH)

    c_state_new = fg_act * c_state + ig_act * (k_scaled @ v.transpose(-1, -2))  # (B, NH, DH, DH)
    n_state_new = fg_act * n_state + ig_act * k_scaled  # (B, NH, DH, 1)

    h_num = q.transpose(-1, -2) @ c_state_new  # (B, NH, 1, DH)

    qn_dotproduct = q.transpose(-1, -2) @ n_state_new  # (B, NH, 1, 1)
    max_val = torch.exp(-m_state_new)  # (B, NH, 1, 1)
    h_denom = torch.maximum(qn_dotproduct.abs(), max_val) + eps
    h = h_num / h_denom  # (B, NH, 1, DH) / (B, NH, 1, 1) = (B, NH, 1, DH)

    return h, (c_state_new, n_state_new, m_state_new)


def chunkwise_simple(
    queries: torch.Tensor,
    keys: torch.Tensor,  # B, NH, S, DH
    values: torch.Tensor,  # B, NH, S, DH
    igate_preact: torch.Tensor,  # B, NH, S
    fgate_preact: torch.Tensor,  # B, NH, S
    initial_C: Optional[torch.Tensor] = None,  # B, NH, DH, DH
    initial_n: Optional[torch.Tensor] = None,  # B, NH, DH
    initial_m: Optional[torch.Tensor] = None,  # B, NH, 1
    chunk_size: int = 64,  # optimize this
    return_last_state: bool = False,
    eps: float = 1e-6,
    **kwargs,
) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
    B, NH, S, DH = queries.shape
    NS, CS = S // chunk_size, chunk_size
    _dtype, _device = queries.dtype, queries.device

    # form chunks
    q = queries.view(B, NH, NS, CS, DH) / math.sqrt(DH)
    k = keys.view(B, NH, NS, CS, DH)
    v = values.view(B, NH, NS, CS, DH)

    # forget gates
    log_fgates = torch.nn.functional.logsigmoid(fgate_preact).view(B, NH, NS, CS)
    log_fgates_acc = log_fgates.cumsum(dim=3)
    igate_preact = igate_preact.view(B, NH, NS, CS)

    loggates = (igate_preact - log_fgates_acc)[:, :, :, :, None]
    m_loc, _ = torch.max(
        loggates + log_fgates_acc[:, :, :, -1, None, None], dim=3, keepdim=True
    )
    loggates = loggates + log_fgates_acc[:, :, :, -1, None, None] - m_loc

    kv = k.transpose(-1, -2) @ (v * (loggates).exp())
    ksum = (k * (loggates).exp()).sum(dim=-2)
    C = torch.zeros((B, NH, NS + 1, DH, DH), device=kv.device, dtype=kv.dtype)
    n = torch.zeros((B, NH, NS + 1, DH), device=kv.device, dtype=kv.dtype)
    if initial_C is not None:
        C[:, :, 0] = initial_C
    if initial_n is not None:
        n[:, :, 0] = initial_n

    m = torch.zeros((B, NH, NS + 1, 1, 1), device=kv.device, dtype=kv.dtype)
    if initial_m is not None:
        m[:, :, 0] = initial_m[:, :, None, None]

    for i in range(1, NS + 1):
        m[:, :, i] = torch.maximum(
            log_fgates_acc[:, :, i - 1, -1, None, None] + m[:, :, i - 1],
            m_loc[:, :, i - 1],
        )
        C[:, :, i] = (
            C[:, :, i - 1].clone()
            * (
                log_fgates_acc[:, :, i - 1, -1, None, None]
                + m[:, :, i - 1]
                - m[:, :, i]
            ).exp()
            + kv[:, :, i - 1] * (m_loc[:, :, i - 1] - m[:, :, i]).exp()
        )
        n[:, :, i] = (
            n[:, :, i - 1].clone()
            * (
                log_fgates_acc[:, :, i - 1, -1, None]
                + m[:, :, i - 1, 0]
                - m[:, :, i, 0]
            ).exp()
            + ksum[:, :, i - 1] * (m_loc[:, :, i - 1, 0] - m[:, :, i, 0]).exp()
        )

    log_fgates_rep = log_fgates_acc[:, :, :, :, None].repeat(1, 1, 1, 1, CS)
    log_fg_matrix = (
        log_fgates_rep
        - log_fgates_rep.transpose(-1, -2)
        - torch.triu(float("inf") * torch.ones([1, 1, 1, CS, CS]).to(q), diagonal=1)
    )

    # gate decay matrix D (combination of forget gate and input gate)
    log_D_matrix = log_fg_matrix + igate_preact[:, :, :, :, None].transpose(
        -2, -1
    )  # (B, NH, NS, CS, CS)
    D_max, _ = torch.max(log_D_matrix, dim=-1, keepdim=True)

    stab = torch.maximum(D_max, m[:, :, :-1, :] + log_fgates_acc[:, :, :, :, None])
    inter_C = (
        q * (m[:, :, :-1, :] + log_fgates_acc[:, :, :, :, None] - stab).exp()
    ) @ C[:, :, :-1]
    inter_n = (
        q * (m[:, :, :-1, :] + log_fgates_acc[:, :, :, :, None] - stab).exp()
    ) @ n[:, :, :-1, :, None]

    # D matrix stabilization
    log_D_matrix_stabilized = log_D_matrix - stab  # (B, NH, NS, CS, CS)
    D_matrix = torch.exp(log_D_matrix_stabilized)  # (B, NH, NS, CS, CS)

    # combination matrix C
    qk_matrix = q @ k.transpose(-2, -1)  # (B, NH, NS, CS, CS)
    E_matrix = qk_matrix * D_matrix  # (B, NH, NS, CS, CS)

    normalizer = torch.maximum(
        (E_matrix.sum(dim=-1, keepdim=True) + inter_n).abs(),
        torch.exp(-stab),
    )  # (B, NH, NS, CS, 1)

    E_matrix_normalized = E_matrix / (normalizer + eps)

    # retrieved values
    intra = E_matrix_normalized @ v  # (B, NH, S, DH)
    inter = inter_C / (normalizer + eps)

    if return_last_state:
        return (intra + inter).view((B, NH, S, DH)), (C[:, :, -1], n[:, :, -1], m[:, :, -1])
    else:
        return (intra + inter).view((B, NH, S, DH))




# Copyright (c) NXAI GmbH and its affiliates 2024
# Korbininan Pöppel

from dataclasses import dataclass, field
from typing import Optional

from ...components.feedforward import FeedForwardConfig
from ..xlstm_block import xLSTMBlock, xLSTMBlockConfig
from .layer import sLSTMLayerConfig


@dataclass
class sLSTMBlockConfig:
    slstm: sLSTMLayerConfig = field(default_factory=sLSTMLayerConfig)
    feedforward: Optional[FeedForwardConfig] = field(default_factory=FeedForwardConfig)

    # we initialize these with None to catch the case where they are not set
    _num_blocks: int = None
    _block_idx: int = None

    def __post_init__(self):
        self.slstm._block_idx = self._block_idx
        self.slstm._num_blocks = self._num_blocks
        self.slstm.__post_init__()
        if self.feedforward is not None:
            self.feedforward.__post_init__()


class sLSTMBlock(xLSTMBlock):
    config_class = sLSTMBlockConfig

    def __init__(self, config: sLSTMBlockConfig):
        super().__init__(
            xLSTMBlockConfig(
                mlstm=None,
                slstm=config.slstm,
                feedforward=config.feedforward,
                _block_idx=config._block_idx,
                _num_blocks=config._num_blocks,
            )
        )



# Copyright (c) NXAI GmbH and its affiliates 2023
# Korbininan Pöppel

import torch
from torch.nn.functional import logsigmoid


def slstm_forward_pointwise(
    Wx: torch.Tensor,  # dim [B, 4*H]
    Ry: torch.Tensor,  # dim [B, 4*H]
    b: torch.Tensor,  # dim [1, 4*H]
    states: torch.Tensor,  # dim [4, B, H]
    constants: dict[str, float],
) -> tuple[
    torch.Tensor,
    torch.Tensor,
]:
    _ = constants
    raw = Wx + Ry + b
    y, c, n, m = torch.unbind(states.view(4, states.shape[1], -1), dim=0)
    # raw = raw.view(-1, 4, -1)
    iraw, fraw, zraw, oraw = torch.unbind(raw.view(raw.shape[0], 4, -1), dim=1)
    # with torch.no_grad():  # THE difference to maxg aka max_gradient (here max / max_static)
    logfplusm = m + logsigmoid(fraw)
    if torch.all(n == 0.0):
        mnew = iraw
    else:
        mnew = torch.max(iraw, logfplusm)
    ogate = torch.sigmoid(oraw)
    igate = torch.exp(iraw - mnew)
    fgate = torch.exp(logfplusm - mnew)
    cnew = fgate * c + igate * torch.tanh(zraw)
    nnew = fgate * n + igate
    ynew = ogate * cnew / nnew

    # shapes ([B,H], [B,H], [B,H]), ([B,H],[B,H],[B,H],[B,H])
    return torch.stack((ynew, cnew, nnew, mnew), dim=0), torch.stack(
        (igate, fgate, zraw, ogate), dim=0
    )



import torch
from torch import nn
from abc import abstractmethod, ABC

def soft_cap(values: torch.Tensor, cap_value: float | torch.Tensor | None) -> torch.Tensor:
    """
    Soft caps a tensor to a value.

    Performs a tanh operation on the logits and scales the result to the cap value. Common technique in attention
    and output language heads to prevent large logits from dominating the softmax. See for example Gemma2:
    https://arxiv.org/abs/2408.00118

    Args:
        values: The tensor to cap.
        cap_value: The value to cap the values to. If None, no cap is applied.

    Returns:
        The capped values.
    """
    if cap_value is None:
        return values
    return cap_value * torch.tanh(values / cap_value)

class NormLayer(nn.Module, ABC):
    """Base class for normalization layers.
    This class contains optional learnable weight and bias parameters.
    
    Args:
        num_features: The number of features in the input tensor.
        eps: A small value to avoid division by zero.
        use_weight: Whether to use a learnable weight.
        use_bias: Whether to use a learnable bias.
        force_float32_reductions: Whether to force float32 reductions.   
    """
    def __init__(
        self,
        num_features: int,
        eps: float = 1e-6,
        use_weight: bool = True,
        use_bias: bool = False,
        force_float32_reductions: bool = True,
    ):
        super().__init__()
        self.num_features = num_features
        self.eps = eps
        self.force_float32_reductions = force_float32_reductions

        if use_weight:
            self.weight = nn.Parameter(torch.ones(num_features))
        else:
            self.weight = None

        if use_bias:
            self.bias = nn.Parameter(torch.zeros(num_features))
        else:
            self.bias = None

    def _apply_weight_bias(self, x: torch.Tensor) -> torch.Tensor:
        if self.weight is not None:
            x = x * self.weight
        if self.bias is not None:
            x = x + self.bias
        return x
    
    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


class RMSNorm(NormLayer):
    """Root mean square normalization layer implementation similar
    to https://pytorch.org/docs/stable/generated/torch.nn.RMSNorm.html. 

    It normalizes the input tensor by the root mean square of the last dimension.

    Args:
        num_features: The number of features in the input tensor.
        eps: A small value to avoid division by zero.
        use_weight: Whether to use a learnable weight.
        use_bias: Whether to use a learnable bias.
        force_float32_reductions: Whether to force float32 reductions.
    """

    def _rms_normalize(self, x: torch.Tensor) -> torch.Tensor:
        # x: (B, ..., S,..., D)
        # apply rms norm over the last dimension, i.e. D dimension
        in_dtype = x.dtype
        if self.force_float32_reductions:
            x = x.float()
        x = x * torch.rsqrt(x.pow(2).mean(dim=-1, keepdim=True) + self.eps)
        return x.to(in_dtype)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (B, ..., S,..., D)
        x = self._rms_normalize(x)
        x = self._apply_weight_bias(x)
        return x

class MultiHeadRMSNorm(RMSNorm):
    """Multi-head version of the RMSNorm layer.
    It normalizes the last dimension of the input tensor by the root mean square.

    The input is assumed to have the shape (B, S, NH, DH), where:
    B: batch size
    S: sequence length
    NH: number of heads
    DH: head dimension
    
    The normalization is applied over the last dimension (DH) of the input tensor.
    Weights and biases are applied after the normalization.

    Args:
        num_heads: The number of heads.
        head_dim: The head dimension.
        eps: A small value to avoid division by zero.
        use_weight: Whether to use a learnable weight.
        use_bias: Whether to use a learnable bias.
        force_float32_reductions: Whether to force float32 reductions
    
    Returns:
        The normalized tensor with the shape (B, S, NH * DH).    
    """
    def __init__(
        self,
        num_heads: int,
        head_dim: int,
        eps: float = 1e-6,
        use_weight: bool = True,
        use_bias: bool = False,
        force_float32_reductions: bool = True,
    ):
        super().__init__(
            num_features=num_heads * head_dim,
            eps=eps,
            use_weight=use_weight,
            use_bias=use_bias,
            force_float32_reductions=force_float32_reductions,
        )
        self.num_heads = num_heads
        self.head_dim = head_dim

    def forward(
        self,
        x: torch.Tensor, # (B, S, NH, DH)
    ) -> torch.Tensor: # (B, S, NH * DH)
        B, S, NH, DH = x.shape
        assert NH == self.num_heads, f"Expected {self.num_heads} heads, got {NH}, input shape: {x.shape}"
        assert DH == self.head_dim, f"Expected {self.head_dim} head dimension, got {DH}, input shape: {x.shape}"

        x = self._rms_normalize(x)
        x = x.reshape(B, S, -1)
        x = self._apply_weight_bias(x)
        return x

class LayerNorm(NormLayer):
    """Layer normalization layer implementation similar to 
    https://pytorch.org/docs/stable/generated/torch.nn.LayerNorm.html.
    
    The layer normalization is applied over the last dimension of the input tensor.

    Args:
        num_features: The number of features in the input tensor.
        eps: A small value to avoid division by zero.
        use_weight: Whether to use a learnable weight.
        use_bias: Whether to use a learnable bias.
        force_float32_reductions: Whether to force float32 reductions.

    Returns:
        The normalized tensor.    
    """

    def _layer_normalize(self, x: torch.Tensor) -> torch.Tensor:
        # x: (B, ..., S,..., D)
        # apply layer norm over the last dimension, i.e. D dimension
        in_dtype = x.dtype
        if self.force_float32_reductions:
            x = x.float()
        x_centered = x - x.mean(dim=-1, keepdim=True)
        y = x_centered * torch.rsqrt(x.var(dim=-1, keepdim=True, unbiased=False) + self.eps)
        return y.to(in_dtype)
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # x: (B, ..., S,..., D)
        x = self._layer_normalize(x)
        x = self._apply_weight_bias(x)
        return x

class MultiHeadLayerNorm(LayerNorm):
    """Multi-head version of the LayerNorm layer.
    
    It normalizes the last dimension of the input tensor.

    The input is assumed to have the shape (B, S, NH, DH), where:
    B: batch size
    S: sequence length
    NH: number of heads
    DH: head dimension

    The normalization is applied over the last dimension (DH) of the input tensor.

    Args:
        num_heads: The number of heads.
        head_dim: The head dimension.
        eps: A small value to avoid division by zero.
        use_weight: Whether to use a learnable weight.
        use_bias: Whether to use a learnable bias.
        force_float32_reductions: Whether to force float32 reductions

    Returns:
        The normalized tensor with the shape (B, S, NH * DH).
    """
    def __init__(
        self,
        num_heads: int,
        head_dim: int,
        eps: float = 1e-6,
        use_weight: bool = True,
        use_bias: bool = False,
        force_float32_reductions: bool = True,
    ):
        super().__init__(
            num_features=num_heads * head_dim,
            eps=eps,
            use_weight=use_weight,
            use_bias=use_bias,
            force_float32_reductions=force_float32_reductions,
        )
        self.num_heads = num_heads
        self.head_dim = head_dim

    def forward(
        self,
        x: torch.Tensor, # (B, S, NH, DH)
    ) -> torch.Tensor: # (B, S, NH * DH)
        B, S, NH, DH = x.shape
        assert NH == self.num_heads, f"Expected {self.num_heads} heads, got {NH}, input shape: {x.shape}"
        assert DH == self.head_dim, f"Expected {self.head_dim} head dimension, got {DH}, input shape: {x.shape}"

        x = self._layer_normalize(x)
        x = x.reshape(B, S, -1)
        x = self._apply_weight_bias(x)
        return x



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
import math

import torch


def bias_linspace_init_(param: torch.Tensor, start: float = 3.4, end: float = 6.0) -> torch.Tensor:
    """Linearly spaced bias init across dimensions."""
    assert param.dim() == 1, f"param must be 1-dimensional (typically a bias), got {param.dim()}"
    n_dims = param.shape[0]
    init_vals = torch.linspace(start, end, n_dims)
    with torch.no_grad():
        param.copy_(init_vals)
    return param


def small_init_init_(param: torch.Tensor, dim: int) -> torch.Tensor:
    """Fills the input Tensor with values according to the method described in Transformers without Tears: Improving
    the Normalization of Self-Attention - Nguyen, T. & Salazar, J. (2019), using a normal distribution.
    Adopted from https://github.com/EleutherAI/gpt-neox/blob/main/megatron/model/init_functions.py.
    """
    std = math.sqrt(2 / (5 * dim))
    torch.nn.init.normal_(param, mean=0.0, std=std)
    return param


def wang_init_(param: torch.Tensor, dim: int, num_blocks: int):
    """Adopted from https://github.com/EleutherAI/gpt-neox/blob/main/megatron/model/init_functions.py."""
    std = 2 / num_blocks / math.sqrt(dim)
    torch.nn.init.normal_(param, mean=0.0, std=std)
    return param



import torch
import pytest

from xlstm.xlstm_large.model import xLSTMLargeConfig, xLSTMLarge
from xlstm.xlstm_large.generate import generate_tokens


def template_test_generation(
    batch_size: int,
    prefill_length: int,
    max_length: int,
    chunkwise_kernel_name: str = "chunkwise--native_autograd",
    sequence_kernel_name: str = "native_sequence__native",
    step_kernel_name: str = "native",
    seed: int = 0,
    vocab_size: int = 512,
    num_blocks: int = 2,
    embedding_dim: int = 256,
    num_heads: int = 4,
    use_torch_compile_model: bool = False,
    use_torch_compile_generate: bool = False,
    device: str = "cuda",
):
    """A generic test template that tests text generation.

    Args:
        batch_size: The batch size.
        prefill_length: The length of the prefill.
        max_length: The maximum length of the generated text.
        chunkwise_kernel_name: The chunkwise kernel to use.
        sequence_kernel_name: The sequence kernel to use.
        step_kernel_name: The step kernel to use.
        seed: The random seed.
        vocab_size: The vocabulary size.
        num_blocks: The number of blocks.
        embedding_dim: The embedding dimension.
        num_heads: The number of heads.
        use_torch_compile_model: Whether to compile the model.
        use_torch_compile_generate: Whether to compile the generation function.
        device: The device to use.
    """
    device_str = device
    device = torch.device(device)
    mlstm_config = xLSTMLargeConfig(
        embedding_dim=embedding_dim,
        num_heads=num_heads,
        num_blocks=num_blocks,
        vocab_size=vocab_size,
        mode="inference",
        chunkwise_kernel=chunkwise_kernel_name,
        sequence_kernel=sequence_kernel_name,
        step_kernel=step_kernel_name,
        return_last_states=True,
    )
    mlstm = xLSTMLarge(mlstm_config)
    if use_torch_compile_model:
        mlstm = torch.compile(mlstm)

    torch.manual_seed(seed)
    mlstm = mlstm.to(device=device)

    if prefill_length > 0:
        prefill_tokens = torch.randint(0, vocab_size, (batch_size, prefill_length)).to(
            device=device
        )
    else:
        prefill_tokens = None
        batch_size = 1

    def llm_forward(tokens, state):
        return mlstm(tokens, state)

    generate_tokens_fn = generate_tokens
    if use_torch_compile_generate:
        generate_tokens_fn = torch.compile(generate_tokens)

    generated_tokens, state = generate_tokens_fn(
        llm_forward=llm_forward,
        prefill_tokens=prefill_tokens,
        max_length=max_length,
        device=device_str,
    )

    assert generated_tokens.shape == (
        batch_size,
        max_length,
    ), f"generated_tokens has wrong shape, got {generated_tokens.shape}"


@pytest.mark.parametrize("batch_size", [1])
@pytest.mark.parametrize("prefill_length", [0])
@pytest.mark.parametrize("max_length", [32, 1, 128])
def test_generate_no_prefill(batch_size: int, prefill_length: int, max_length: int):
    template_test_generation(
        batch_size=batch_size,
        prefill_length=prefill_length,
        max_length=max_length,
    )


@pytest.mark.parametrize(
    "batch_size, prefill_length, max_length",
    [
        [2, 128, 128],
        [2, 150, 128],
    ],
)
@pytest.mark.parametrize(
    "chunkwise_kernel, sequence_kernel, step_kernel",
    [
        [
            "chunkwise--triton_limit_chunk",
            "native_sequence__triton_step_fused",
            "triton_fused",
        ],
        [
            "chunkwise--triton_xl_chunk",
            "native_sequence__triton_step_fused",
            "triton_fused",
        ],
    ],
)
def test_generate_prefill(
    batch_size: int,
    prefill_length: int,
    max_length: int,
    chunkwise_kernel: str,
    sequence_kernel: str,
    step_kernel: str,
):
    """Tests text generation with prefilling."""
    template_test_generation(
        chunkwise_kernel_name=chunkwise_kernel,
        sequence_kernel_name=sequence_kernel,
        step_kernel_name=step_kernel,
        batch_size=batch_size,
        prefill_length=prefill_length,
        max_length=max_length,
    )


@pytest.mark.parametrize(
    "batch_size, prefill_length, max_length",
    [
        [2, 128, 128],
    ],
)
@pytest.mark.parametrize(
    "chunkwise_kernel, sequence_kernel, step_kernel",
    [
        [
            "chunkwise--triton_limit_chunk",
            "native_sequence__triton_step_fused",
            "triton_fused",
        ],
        [
            "chunkwise--triton_xl_chunk",
            "native_sequence__triton_step_fused",
            "triton_fused",
        ],
    ],
)
@pytest.mark.parametrize(
    "compile_model, compile_generate", [[False, False], [True, False]]
)
def test_generate_compile(
    batch_size: int,
    prefill_length: int,
    max_length: int,
    chunkwise_kernel: str,
    sequence_kernel: str,
    step_kernel: str,
    compile_model: bool,
    compile_generate: bool,
):
    """Tests text generation with compilation."""
    template_test_generation(
        chunkwise_kernel_name=chunkwise_kernel,
        sequence_kernel_name=sequence_kernel,
        step_kernel_name=step_kernel,
        batch_size=batch_size,
        prefill_length=prefill_length,
        max_length=max_length,
        use_torch_compile_model=compile_model,
        use_torch_compile_generate=compile_generate,
    )
    torch.compiler.reset()

@pytest.mark.parametrize(
    "batch_size, prefill_length, max_length",
    [
        [2, 128, 128],
    ],
)
def test_model_generate(
    batch_size: int,
    prefill_length: int,
    max_length: int,
    chunkwise_kernel_name: str = "chunkwise--native_autograd",
    sequence_kernel_name: str = "native_sequence__native",
    step_kernel_name: str = "native",
    seed: int = 0,
    vocab_size: int = 512,
    num_blocks: int = 2,
    embedding_dim: int = 256,
    num_heads: int = 4,
    device: str = "cuda",
):
    """Tests calling the generation from the model."""
    torch.manual_seed(seed)
    device = torch.device(device)
    mlstm_config = xLSTMLargeConfig(
        embedding_dim=embedding_dim,
        num_heads=num_heads,
        num_blocks=num_blocks,
        vocab_size=vocab_size,
        mode="inference",
        chunkwise_kernel=chunkwise_kernel_name,
        sequence_kernel=sequence_kernel_name,
        step_kernel=step_kernel_name,
        return_last_states=True,
    )
    mlstm = xLSTMLarge(mlstm_config).to(device=device)

    if prefill_length > 0:
        prefill_tokens = torch.randint(0, vocab_size, (batch_size, prefill_length)).to(
            device=device
        )
    else:
        prefill_tokens = None
        batch_size = 1

    generated_tokens, state = mlstm.generate(
        prefill_tokens=prefill_tokens,
        max_length=max_length,
    )

    assert generated_tokens.shape == (
        batch_size,
        max_length,
    ), f"generated_tokens has wrong shape, got {generated_tokens.shape}"



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck, Korbinian Pöppel
import torch
import torch.nn.functional as F
from torch import nn


class LayerNorm(nn.Module):
    """LayerNorm but with an optional bias. PyTorch doesn't support simply bias=False."""

    def __init__(
        self,
        ndim: int = -1,
        weight: bool = True,
        bias: bool = False,
        eps: float = 1e-5,
        residual_weight: bool = True,
    ):
        super().__init__()
        self.weight = nn.Parameter(torch.zeros(ndim)) if weight else None
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None
        self.eps = eps
        self.residual_weight = residual_weight
        self.ndim = ndim
        self.reset_parameters()

    @property
    def weight_proxy(self) -> torch.Tensor:
        if self.weight is None:
            return None
        if self.residual_weight:
            return 1.0 + self.weight
        else:
            return self.weight

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.layer_norm(
            input, normalized_shape=(self.ndim,), weight=self.weight_proxy, bias=self.bias, eps=self.eps
        )

    def reset_parameters(self):
        if self.weight_proxy is not None:
            if self.residual_weight:
                nn.init.zeros_(self.weight)
            else:
                nn.init.ones_(self.weight)
        if self.bias is not None:
            nn.init.zeros_(self.bias)


class MultiHeadLayerNorm(LayerNorm):

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        assert input.dim() == 4, "Input must be 4D tensor (B, NH, S, DH)"
        B, NH, S, DH = input.shape

        gn_in_1 = input.transpose(1, 2)  # (B, S, NH, DH)
        gn_in_2 = gn_in_1.reshape(B * S, NH * DH)  # (B * S, NH * DH)
        out = F.group_norm(
            gn_in_2,
            num_groups=NH,
            weight=self.weight_proxy,
            bias=self.bias,
            eps=self.eps,
        )
        # (B * S), (NH * DH) -> (B, S, NH, DH) -> (B, NH, S, DH)
        out = out.view(B, S, NH, DH).transpose(1, 2)
        return out



import pytest
from xlstm.xlstm_large.model import xLSTMLargeConfig, xLSTMLarge
from xlstm.xlstm_large.utils import convert_single_weights_to_fused_weights
import copy
import torch
import numpy as np

@pytest.mark.parametrize("use_bias", [True, False])
def test_single_weights_to_fused_weights(use_bias: bool, device="cpu"):
    """Tests weight conversion of the model from single to fused mode."""
    mlstm_config = xLSTMLargeConfig(
        embedding_dim=512,
        num_heads=4,
        num_blocks=6,
        vocab_size=2048,
        return_last_states=True,
        mode="inference",
        chunkwise_kernel="chunkwise--native_custbw",
        sequence_kernel="native_sequence__native",
        step_kernel="native",
        weight_mode="single",
        use_bias=use_bias,
    )

    mlstm_single = xLSTMLarge(mlstm_config)

    mlstm_fused_config = copy.deepcopy(mlstm_config)
    mlstm_fused_config.weight_mode = "fused"

    mlstm_fused = xLSTMLarge(mlstm_fused_config)

    # copy weights from single to fused model
    new_state_dict = convert_single_weights_to_fused_weights(mlstm_single.state_dict())

    mlstm_fused.load_state_dict(new_state_dict)

    device = torch.device(device)

    mlstm_single.to(device)
    mlstm_fused.to(device)

    inputs = torch.randint(0, mlstm_config.vocab_size, (3, 256)).to(device)

    single_output = mlstm_single(inputs)
    fused_output = mlstm_fused(inputs)

    single_out = single_output[0].cpu().detach().numpy()
    fused_out = fused_output[0].cpu().detach().numpy()

    assert single_out.shape == fused_out.shape
    assert not (single_out == 0.0).all()

    np.testing.assert_allclose(
        actual=fused_out, desired=single_out, rtol=1e-4, atol=1e-4
    )



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from dataclasses import dataclass, field

from ..xlstm_block import xLSTMBlock, xLSTMBlockConfig
from .layer import mLSTMLayerConfig


@dataclass
class mLSTMBlockConfig:
    mlstm: mLSTMLayerConfig = field(default_factory=mLSTMLayerConfig)

    # we initialize these with None to catch the case where they are not set
    _num_blocks: int = None
    _block_idx: int = None

    def __post_init__(self):
        self.mlstm._num_blocks = self._num_blocks
        self.mlstm.__post_init__()


class mLSTMBlock(xLSTMBlock):
    config_class = mLSTMBlockConfig

    def __init__(self, config: mLSTMBlockConfig) -> None:
        super().__init__(
            config=xLSTMBlockConfig(
                mlstm=config.mlstm,
                slstm=None,
                feedforward=None,
                _num_blocks=config._num_blocks,
                _block_idx=config._block_idx,
            )
        )



__version__ = "2.0.1"

from .blocks.mlstm.block import mLSTMBlock, mLSTMBlockConfig
from .blocks.mlstm.layer import mLSTMLayer, mLSTMLayerConfig
from .blocks.slstm.block import sLSTMBlock, sLSTMBlockConfig
from .blocks.slstm.layer import sLSTMLayer, sLSTMLayerConfig
from .components.feedforward import FeedForwardConfig, GatedFeedForward
from .xlstm_block_stack import xLSTMBlockStack, xLSTMBlockStackConfig
from .xlstm_lm_model import xLSTMLMModel, xLSTMLMModelConfig



from typing import Callable

import torch
from torch.profiler import record_function

LLMStateType = dict

LLMForwardFnType = Callable[
    [torch.Tensor, LLMStateType | None], tuple[torch.Tensor, LLMStateType]
]


def greedy_sampling(logits: torch.Tensor) -> torch.Tensor:
    """Takes the argmax of the logits. Actually no sampling is done here."""
    return torch.argmax(logits, dim=-1)

# Ths registry is used to map the sampling function name to the actual function
_sampling_fn_registry = {
    "greedy": greedy_sampling,
}


def get_sampling_fn(sampling_fn_name: str) -> Callable[[torch.Tensor], torch.Tensor]:
    return _sampling_fn_registry[sampling_fn_name]


@torch.no_grad
def generate_tokens(
    llm_forward: LLMForwardFnType,
    prefill_tokens: torch.Tensor | None = None,
    max_length: int = 128,
    token_sample_fn: Callable[[torch.Tensor], torch.Tensor] = greedy_sampling,
    bos_token_id: int = 0,
    state: LLMStateType | None = None,
    batch_size_no_prefill: int = 1,
    generated_tokens: torch.Tensor | None = None,
    device: str = "cuda",
) -> tuple[torch.Tensor, LLMStateType]:
    """A simple function to generate tokens from a language model.
    It generates exactly max_length tokens.

    Args:
        llm_forward: The forward function of the language model.
        prefill_tokens: The tokens to prefill the model with.
        max_length: The maximum length of the generated tokens.
        token_sample_fn: The function to sample the next token.
        bos_token_id: The id of the beginning of sequence token.
        state: The state of the language model.
        batch_size_no_prefill: The batch size used for generation if no prefill_tokens are given.
        generated_tokens: The generated tokens tensor.
        device: The device to use.

    Returns:
        generated_tokens: The generated tokens tensor.
        state: The state of the language model after generation.
    """

    if prefill_tokens is None:
        prefill_tokens = torch.full((batch_size_no_prefill, 1), fill_value=bos_token_id, dtype=torch.long, device=device)
    
    if prefill_tokens.ndim == 1:
        prefill_tokens = prefill_tokens[:, None]

    batch_size = prefill_tokens.size(0)

    if max_length > 0:
        # init the generated tokens tensors
        # we add 1 to the max_length to account for the BOS token
        if generated_tokens is None:
            generated_tokens = torch.empty(
                (batch_size, max_length), dtype=torch.long, device=device
            )
        else:
            assert generated_tokens.shape == (batch_size, max_length), (
                f"Generated tokens shape: {tuple(generated_tokens.shape)}, "
                f"expected {(batch_size, max_length)}"
            )

        last_token = prefill_tokens
        for i in range(max_length):
            with record_function(f"generate_tokens_step_{i}"):
                logits, state = llm_forward(last_token, state)
                next_token = token_sample_fn(logits[:, -1:])
                generated_tokens[:, i:i+1] = next_token
                last_token = next_token
    else:
        # we only return the state
        generated_tokens = None

    return generated_tokens, state


# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Literal, Optional, Union

import torch
from torch import nn

from .blocks.mlstm.block import mLSTMBlock, mLSTMBlockConfig
from .blocks.slstm.block import sLSTMBlock, sLSTMBlockConfig
from .components.ln import LayerNorm


@dataclass
class xLSTMBlockStackConfig:
    mlstm_block: Optional[mLSTMBlockConfig] = None
    slstm_block: Optional[sLSTMBlockConfig] = None

    context_length: int = -1
    num_blocks: int = 1
    embedding_dim: int = 128
    add_post_blocks_norm: bool = True
    bias: bool = False
    dropout: float = 0.0

    # The block indices at which sLSTM blocks are placed.
    # Indexing starts from 0.
    slstm_at: Union[list[int], Literal["all"]] = field(default_factory=list)

    # _block_map is a string that specifies which block is used at which position
    # 0: use the mLSTM block
    # 1: use the sLSTM block
    _block_map: str = None

    @property
    def block_map(self) -> list[int]:
        return list(map(int, self._block_map.split(",")))

    def _create_block_map(self) -> str:
        """Creates the block map, that specifies which block is used at which position."""
        block_map = [0] * self.num_blocks

        for slstm_position_idx in self.slstm_at:
            assert slstm_position_idx < self.num_blocks, f"Invalid slstm position {slstm_position_idx}"
            block_map[slstm_position_idx] = 1

        block_map_str = ",".join(map(str, block_map))

        return block_map_str

    def __post_init__(self):
        if self.mlstm_block is None:
            self.slstm_at = "all"
        if self.slstm_at == "all":
            self.slstm_at = list(range(self.num_blocks))

        if self.mlstm_block is not None:
            self.mlstm_block.mlstm.embedding_dim = self.embedding_dim
            self.mlstm_block.mlstm.bias = self.bias
            self.mlstm_block.mlstm.dropout = self.dropout
            self.mlstm_block.mlstm.context_length = self.context_length
            
            self.mlstm_block._num_blocks = self.num_blocks
            # call post init, for setting inner_embedding_dim
            self.mlstm_block.__post_init__()

        if self.slstm_block is not None:
            self.slstm_block.slstm.dropout = self.dropout
            self.slstm_block.slstm.embedding_dim = self.embedding_dim
            self.slstm_block._num_blocks = self.num_blocks
            self.slstm_block.__post_init__()

        self._block_map = self._create_block_map()


class xLSTMBlockStack(nn.Module):
    config_class = xLSTMBlockStackConfig

    def __init__(self, config: xLSTMBlockStackConfig):
        super().__init__()
        self.config = config

        self.blocks = self._create_blocks(config=config)
        if config.add_post_blocks_norm:
            self.post_blocks_norm = LayerNorm(ndim=config.embedding_dim)
        else:
            self.post_blocks_norm = nn.Identity()

    def _create_blocks(self, config: xLSTMBlockStackConfig):

        blocks = []
        for block_idx, block_type_int in enumerate(config.block_map):
            if block_type_int == 0:
                config = deepcopy(self.config.mlstm_block)
                if hasattr(config, "_block_idx"):
                    config._block_idx = block_idx
                    config.__post_init__()
                blocks.append(mLSTMBlock(config=config))
            elif block_type_int == 1:
                config = deepcopy(self.config.slstm_block)
                if hasattr(config, "_block_idx"):
                    config._block_idx = block_idx
                    config.__post_init__()
                blocks.append(sLSTMBlock(config=config))
            else:
                raise ValueError(f"Invalid block type {block_type_int}")

        return nn.ModuleList(blocks)

    def reset_parameters(self) -> None:
        for block in self.blocks:
            block.reset_parameters()
        if not isinstance(self.post_blocks_norm, nn.Identity):
            self.post_blocks_norm.reset_parameters()

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:

        for block in self.blocks:
            x = block(x, **kwargs)

        x = self.post_blocks_norm(x)

        return x

    def step(
        self, x: torch.Tensor, state: dict[str, dict[str, tuple[torch.Tensor, ...]]] = None
    ) -> tuple[torch.Tensor, dict[str, dict[str, tuple[torch.Tensor, ...]]]]:
        if state is None:
            state = {}

        for block_idx, block in enumerate(self.blocks):
            x, state[f"block_{block_idx}"] = block.step(x, **state.get(f"block_{block_idx}", {}))

        x = self.post_blocks_norm(x)

        return x, state



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from dataclasses import dataclass
from typing import Callable, Literal

import torch
from torch import nn

from ..utils import UpProjConfigMixin
from .init import small_init_init_, wang_init_

_act_fn_registry = {
    "gelu": nn.functional.gelu,
    "relu": nn.functional.relu,
    "relu^2": lambda x: torch.square(nn.functional.relu(x)),
    "sigmoid": nn.functional.sigmoid,
    "swish": nn.functional.silu,
    "selu": nn.functional.selu,
}


def get_act_fn(act_fn_name: str) -> Callable[[torch.Tensor], torch.Tensor]:
    if act_fn_name in _act_fn_registry:
        return _act_fn_registry[act_fn_name]
    else:
        assert (
            False
        ), f'Unknown activation function name "{act_fn_name}". Available activation functions are: {str(_act_fn_registry.keys())}'


@dataclass
class FeedForwardConfig(UpProjConfigMixin):
    proj_factor: float = 1.3
    act_fn: str = "gelu"
    embedding_dim: int = -1
    dropout: float = 0.0
    bias: bool = False
    ff_type: Literal["ffn_gated"] = "ffn_gated"

    _num_blocks: int = 1

    def __post_init__(self):
        self._set_proj_up_dim(embedding_dim=self.embedding_dim)
        assert (
            self.act_fn in _act_fn_registry
        ), f"Unknown activation function {self.act_fn}"


class GatedFeedForward(nn.Module):
    config_class = FeedForwardConfig

    def __init__(self, config: FeedForwardConfig):
        super().__init__()
        self.config = config

        self.proj_up = nn.Linear(
            in_features=self.config.embedding_dim,
            out_features=2 * self.config._proj_up_dim,
            bias=self.config.bias,
        )
        self.proj_down = nn.Linear(
            in_features=self.config._proj_up_dim,
            out_features=self.config.embedding_dim,
            bias=self.config.bias,
        )

        self.act_fn = get_act_fn(self.config.act_fn)

        self.dropout = nn.Dropout(self.config.dropout)
        self.reset_parameters()

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        gate_preact, up_proj = self.proj_up(x).split(self.config._proj_up_dim, dim=-1)
        x = self.dropout(self.proj_down(self.act_fn(gate_preact) * up_proj))
        return x

    def reset_parameters(self):
        small_init_init_(self.proj_up.weight, dim=self.config.embedding_dim)
        if self.proj_up.bias is not None:
            nn.init.zeros_(self.proj_up.bias)
        wang_init_(
            self.proj_down.weight,
            dim=self.config.embedding_dim,
            num_blocks=self.config._num_blocks,
        )
        if self.proj_down.bias is not None:
            nn.init.zeros_(self.proj_down.bias)


def create_feedforward(config: FeedForwardConfig) -> nn.Module:
    if config.ff_type == "ffn_gated":
        return GatedFeedForward(config)
    else:
        raise ValueError(f"Unknown feedforward type {config.ff_type}")



from pathlib import Path

from omegaconf import OmegaConf
from safetensors.torch import load_file

from .model import xLSTMLarge, xLSTMLargeConfig


def load_from_pretrained(
    checkpoint_path: str | Path,
    return_last_states: bool | None = None,
    chunkwise_kernel_name: str | None = None,
    sequence_kernel_name: str | None = None,
    step_kernel_name: str | None = None,
    backend_mode: str | None = "inference",
    chunk_size: int | None = None,
) -> xLSTMLarge:
    """
    Load a mLSTM model from a checkpoint.

    Args:
        checkpoint_path: The path to the checkpoint.
        return_last_states: Whether to return the last states. Overwrites the value from loaded config.
        chunkwise_kernel_name: The chunkwise kernel to use. Overwrites the value from loaded config.
        sequence_kernel_name: The sequence kernel to use. Overwrites the value from loaded config.
        step_kernel_name: The step kernel to use. Overwrites the value from loaded config.
        backend_mode: The backend mode to use. Overwrites the value from loaded config.
        chunk_size: The chunk size to use. Overwrites the value from loaded config.

    Returns:
        mLSTM: The loaded mLSTM model.
    """

    checkpoint_path = Path(checkpoint_path)
    non_sharded_path = checkpoint_path / "model.safetensors"
    if non_sharded_path.exists():
        state_dict = load_file(non_sharded_path)
    else:
        n = 0
        sharded_path = checkpoint_path / f"model_{n}.safetensors"
        state_dict = {}
        while sharded_path.exists():
            state_dict.update(load_file(sharded_path))
            n += 1
            sharded_path = checkpoint_path / f"model_{n}.safetensors"
    config = OmegaConf.load(checkpoint_path / "config.yaml")

    mlstm_config = xLSTMLargeConfig(**config)
    # Note: The default weight mode is single.
    # For fused weight mode convert the weights using convert_single_weights_to_fused_weights.
    mlstm_config.weight_mode = "single"
    if chunkwise_kernel_name is not None:
        mlstm_config.chunkwise_kernel = chunkwise_kernel_name
    if sequence_kernel_name is not None:
        mlstm_config.sequence_kernel = sequence_kernel_name
    if step_kernel_name is not None:
        mlstm_config.step_kernel = step_kernel_name
    if backend_mode is not None:
        mlstm_config.mode = backend_mode
    if return_last_states is not None:
        mlstm_config.return_last_states = return_last_states
    if chunk_size is not None:
        mlstm_config.chunk_size = chunk_size

    mlstm = xLSTMLarge(mlstm_config)
    mlstm.load_state_dict(state_dict)

    return mlstm


import torch
import re


def round_up_to_next_multiple_of(x: int, multiple_of: int) -> int:
    """Rounds up x to the next multiple of multiple_of."""
    return int(((x + multiple_of - 1) // multiple_of) * multiple_of)


def convert_single_weights_to_fused_weights(
    single_weight_state_dict: dict[str, torch.Tensor],
) -> dict[str, torch.Tensor]:
    def get_matching_keys_for(regex: str, keys: list[str]):
        matching_keys = []
        for key in keys:
            if re.match(regex, key):
                matching_keys.append(key)
        return matching_keys

    def concat_weights_and_biases(
        state_dict: dict[str, torch.Tensor],
        weights_and_biases_regex: str,
        first_key_ending: str,
        new_key_ending: str,
    ):
        # fuse qkv o weights
        wb_keys = get_matching_keys_for(weights_and_biases_regex, block_keys)
        if len(wb_keys) == 0:
            return state_dict
        # qkv o are sorted here
        tensors_to_fuse = []
        for key in wb_keys:
            tensors_to_fuse.append(state_dict.pop(key))

        # add qkv o weight to state dict
        fused_weight = torch.cat(tensors_to_fuse, dim=0)
        fused_weight_key = wb_keys[0].replace(
            first_key_ending, new_key_ending
        )

        state_dict.update({fused_weight_key: fused_weight})
        return state_dict

    def convert_mlstm_layer_weights_(
        state_dict: dict[str, torch.Tensor], block_keys: list[str]
    ):
        """Modifies the state dict in place."""

        # fuse qkv o weights
        state_dict = concat_weights_and_biases(
            state_dict,
            ".*(q|k|v|ogate_preact).weight",
            ".q.weight",
            ".qkv_opreact.weight",
        )

        # fuse qkv o biases
        state_dict = concat_weights_and_biases(
            state_dict,
            ".*(q|k|v|ogate_preact).bias",
            ".q.bias",
            ".qkv_opreact.bias",
        )

        # fuse if gate weights
        state_dict = concat_weights_and_biases(
            state_dict,
            ".*(i|f)gate_preact.weight",
            ".igate_preact.weight",
            ".ifgate_preact.weight",
        )

        # fuse if gate biases
        state_dict = concat_weights_and_biases(
            state_dict,
            ".*(i|f)gate_preact.bias",
            ".igate_preact.bias",
            ".ifgate_preact.bias",
        )

        return state_dict

    def convert_feedforward_weights(
        state_dict: dict[str, torch.Tensor], block_keys: list[str]
    ):
        """Modifies the state dict in place."""
        # fuse feedforward weights
        state_dict = concat_weights_and_biases(
            state_dict,
            ".*(proj_up_gate|proj_up).weight",
            ".proj_up_gate.weight",
            ".proj_up_gate_z.weight",
        )

        # fuse feedforward biases
        state_dict = concat_weights_and_biases(
            state_dict,
            ".*(proj_up_gate|proj_up).bias",
            ".proj_up_gate.bias",
            ".proj_up_gate_z.bias",
        )

        return state_dict

    # iterate over blocks and convert weights blockwise
    block_idx = 0
    blocks_key_str = "backbone.blocks.{block_idx}"
    state_dict_keys = list(single_weight_state_dict.keys())
    while True:
        block_key_str = blocks_key_str.format(block_idx=block_idx)
        block_keys = get_matching_keys_for(block_key_str, state_dict_keys)
        if len(block_keys) == 0:
            break

        single_weight_state_dict = convert_mlstm_layer_weights_(
            single_weight_state_dict, block_keys
        )

        single_weight_state_dict = convert_feedforward_weights(
            single_weight_state_dict, block_keys
        )

        block_idx += 1

    return single_weight_state_dict



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
from dataclasses import dataclass
from typing import Optional

import torch
from torch import nn

from ..components.feedforward import FeedForwardConfig, create_feedforward
from ..components.ln import LayerNorm
from .mlstm.layer import mLSTMLayer, mLSTMLayerConfig
from .slstm.layer import sLSTMLayer, sLSTMLayerConfig



@dataclass
class xLSTMBlockConfig:
    mlstm: Optional[mLSTMLayerConfig] = None
    slstm: Optional[sLSTMLayerConfig] = None

    feedforward: Optional[FeedForwardConfig] = None
    
    # we initialize these with None to catch the case where they are not set
    _num_blocks: int = None
    _block_idx: int = None

    def __post_init__(self):
        assert self.mlstm is not None or self.slstm is not None, "Either mlstm or slstm must be provided"
        assert self.mlstm is None or self.slstm is None, "Only one of mlstm or slstm can be provided"
        embedding_dim = self.mlstm.embedding_dim if self.mlstm is not None else self.slstm.embedding_dim
        if self.mlstm:
            self.mlstm._num_blocks = self._num_blocks
            self.mlstm._block_idx = self._block_idx
        if self.slstm:
            self.slstm._num_blocks = self._num_blocks
            self.slstm._block_idx = self._block_idx
        if self.feedforward:
            self.feedforward.embedding_dim = embedding_dim
            self.feedforward._num_blocks = self._num_blocks
            self.feedforward.__post_init__()


class xLSTMBlock(nn.Module):
    """An xLSTM block can be either an sLSTM Block or an mLSTM Block.

    It contains the pre-LayerNorms and the skip connections.
    """

    config_class = xLSTMBlockConfig

    def __init__(self, config: xLSTMBlockConfig) -> None:
        super().__init__()
        self.config = config
        embedding_dim = (
            self.config.mlstm.embedding_dim if self.config.mlstm is not None else self.config.slstm.embedding_dim
        )

        self.xlstm_norm = LayerNorm(ndim=embedding_dim, weight=True, bias=False)

        if self.config.mlstm is not None:
            self.xlstm = mLSTMLayer(config=self.config.mlstm)
        elif self.config.slstm is not None:
            self.xlstm = sLSTMLayer(config=self.config.slstm)
        else:
            raise ValueError("Either mlstm or slstm must be provided")

        if self.config.feedforward is not None:
            self.ffn_norm = LayerNorm(ndim=self.config.feedforward.embedding_dim, weight=True, bias=False)
            self.ffn = create_feedforward(config=self.config.feedforward)
        else:
            self.ffn_norm = None
            self.ffn = None

        self.reset_parameters()

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        x = x + self.xlstm(self.xlstm_norm(x), **kwargs)
        if self.ffn is not None:
            x = x + self.ffn(self.ffn_norm(x), **kwargs)
        return x

    def step(self, x: torch.Tensor, **kwargs) -> tuple[torch.Tensor, dict[str, tuple[torch.Tensor, ...]]]:
        x_xlstm, xlstm_state = self.xlstm.step(self.xlstm_norm(x), **kwargs)
        x = x + x_xlstm
        if self.ffn is not None:
            x = x + self.ffn(self.ffn_norm(x), **kwargs)
        return x, xlstm_state

    def reset_parameters(self) -> None:
        self.xlstm.reset_parameters()
        self.xlstm_norm.reset_parameters()
        if self.ffn is not None:
            self.ffn.reset_parameters()
            self.ffn_norm.reset_parameters()



# Copyright (c) NXAI GmbH and its affiliates 2024
# Maximilian Beck
import math
from abc import ABC
from dataclasses import dataclass
from typing import Sequence

from torch import nn


@dataclass
class UpProjConfigMixin:
    proj_factor: float = None  # will be overridden by subclasses
    round_proj_up_dim_up: bool = True
    round_proj_up_to_multiple_of: int = 64

    # internal
    _proj_up_dim: int = None  # will be computed from embedding_dim and proj_factor

    def _set_proj_up_dim(self, embedding_dim: int) -> None:
        if self.proj_factor is not None and embedding_dim is not None:
            proj_up_dim = self.proj_factor * embedding_dim
            multiple_of_multiplier = proj_up_dim / self.round_proj_up_to_multiple_of
            if self.round_proj_up_dim_up:
                multiple_of_multiplier = math.ceil(multiple_of_multiplier)
            else:
                multiple_of_multiplier = math.floor(multiple_of_multiplier)

            self._proj_up_dim = int(multiple_of_multiplier * self.round_proj_up_to_multiple_of)


class WeightDecayOptimGroupMixin(nn.Module, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_weight_decay_optim_groups(self, **kwargs) -> tuple[Sequence[nn.Parameter], Sequence[nn.Parameter]]:
        """Return a tuple of two sequences, one for parameters with weight decay and one for parameters without weight decay.
        Performs checks to ensure that each parameter is only in one of the two sequences.
        """
        weight_decay, no_weight_decay = self._create_weight_decay_optim_groups(**kwargs)

        # Check that parameters have been assigned correctly.
        # Each parameter can only be in one optim group.
        intersection_params = set(weight_decay).intersection(set(no_weight_decay))
        assert (
            len(intersection_params) == 0
        ), f"parameters {[pn for pn, p in self.named_parameters() if p in intersection_params]} made it into both decay/no_decay sets!"

        union_params = set(weight_decay).union(set(no_weight_decay))
        param_dict = {pn: p for pn, p in self.named_parameters()}
        unassigned_params = set(param_dict.values()) - union_params
        unassigned_params = [up for up in unassigned_params if not hasattr(up, "requires_grad") or up.requires_grad]
        # We have parameters that were not assigned to either weight decay or no weight decay.
        # Find the parameter names and raise an error.
        assert (
            len(unassigned_params) == 0
        ), f"Parameters {[pn for pn, p in self.named_parameters() if all([p is not q for q in unassigned_params])]} were not separated into either decay/no_decay set!"

        return weight_decay, no_weight_decay

    def get_weight_decay_optim_group_param_names(self, **kwargs) -> tuple[Sequence[str], Sequence[str]]:
        """Return a tuple of two sequences, one for parameter names with weight decay and one for parameter names without weight decay.
        Performs checks to ensure that each parameter is only in one of the two sequences.
        """

        def _is_in_sequence(param: nn.Parameter, sequence: Sequence[nn.Parameter]) -> bool:
            for p in sequence:
                if param is p:
                    return True
            return False

        weight_decay, no_weight_decay = self.get_weight_decay_optim_groups(**kwargs)
        names_weight_decay = [pn for pn, p in self.named_parameters() if _is_in_sequence(p, weight_decay)]
        names_no_weight_decay = [pn for pn, p in self.named_parameters() if _is_in_sequence(p, no_weight_decay)]
        return names_weight_decay, names_no_weight_decay

    def _create_weight_decay_optim_groups(self, **kwargs) -> tuple[Sequence[nn.Parameter], Sequence[nn.Parameter]]:
        """Return a tuple of two sequences, one for parameters with weight decay and one for parameters without weight decay.
        Default separation:
        - weight decay: all parameters which have > 1 dimensions.
        - no weight decay: all parameters which have = 1 dimension, e.g. biases.
        """

        decay = set()
        no_decay = set()
        for name, param in self.named_parameters():
            if param.requires_grad:
                # TODO: this is a hack based on module naming. It fixes the issue with the previous heuristic using
                #  parameter dim, which fails for FSDP because it flattens parameters so ndim is always 1.
                if 'weight' in name and 'norm' not in name:
                    decay.add(param)
                else:
                    no_decay.add(param)

        return tuple(decay), tuple(no_decay)

    def _get_weight_decay_optim_groups_for_modules(
        self, modules: list["WeightDecayOptimGroupMixin"], **kwargs
    ) -> tuple[Sequence[nn.Parameter], Sequence[nn.Parameter]]:
        weight_decay, no_weight_decay = (), ()
        for module in modules:
            wd, nwd = module.get_weight_decay_optim_groups(**kwargs)
            weight_decay += wd
            no_weight_decay += nwd
        return weight_decay, no_weight_decay
